
# Generated by OTF2 Template Engine

from collections import defaultdict, OrderedDict
from itertools import chain
from functools import wraps

import _otf2

from . import definitions
from .error import TraceReaderError, TraceWriterError
from .enums import Undefined, GroupType
from .definitions import ClockProperties


def _use_doc_from(doc):
    def decorator(func):
        doc_str = "Returns the {} definition with matching arguments\n".format(doc.__name__)
        if isinstance(doc, definitions._RefDefinition):
            doc_str += "If there is no matching definition, one will be created.\n"
        doc_str += "\n"
        doc_str += doc.__doc__

        func.__doc__ = doc_str
        return func
    return decorator


class _Registry(object):
    def __init__(self, definition_registry, default_type=None):
        self._definition_registry = definition_registry
        self._default_type = default_type
        self._extra_callbacks = {}

    def _callback_gen(self, definition_type=None):
        if definition_type is None:
            definition_type = self._default_type

        def callback(definition_registry, *args):
            definition = definition_type._construct(definition_registry, *args)
            try:
                self._register(definition)
            except KeyError:
                raise TraceReaderError("Duplicate definition: {}".format(definition))
        self._extra_callbacks[definition_type] = callback
        return callback

    def create(self, *args, **kwargs):
        t = kwargs.pop('_type', self._default_type)
        obj = t(self, *args, **kwargs)
        self._register(obj)
        return obj


class _ListRegistry(_Registry):
    def __init__(self, definition_registry, default_type=None):
        self._elements = []
        super(_ListRegistry, self).__init__(definition_registry, default_type)

    def _register(self, obj):
        self._elements.append(obj)

    def __iter__(self):
        """ Do not try to add elements during iteration """
        for elem in self._elements:
            yield elem


class _RefRegistry(_Registry):
    """
    All objects recorded by this registry must have:
    - a ref (id) of type int as first constructor parameter.
      This must not be changed once an event with that ref is written in the trace
    - a name of type str as a second constructor parameter.
      If this is changed, the _elements_by_name must be updated
     """
    def __init__(self, definition_registry, undefined_ref, default_type=None):
        self._elements_by_key = dict()
        self._elements_by_ref = OrderedDict({undefined_ref.value: None})

        self._ref = -1
        self._extra_callbacks = dict()
        super(_RefRegistry, self).__init__(definition_registry, default_type)

    def _next_ref(self):
        self._ref += 1
        return self._ref

    def create(self, *args, **kwargs):
        ref = self._next_ref()
        return super(_RefRegistry, self).create(ref, *args, **kwargs)

    def get(self, *args, **kwargs):
        t = kwargs.pop('_type', self._default_type)
        key = t._new_key(self, *args, **kwargs)
        try:
            return self._elements_by_key[key]
        except KeyError:
            pass
        return self.create(*args, _type=t, **kwargs)

    def _register(self, obj):
        if obj._ref in self._elements_by_ref:
            raise KeyError("Trying to register definition with existing reference.")
        self._elements_by_key[obj._key] = obj
        self._elements_by_ref[obj._ref] = obj
        # Maybe redundant, don't care
        self._ref = max(self._ref, obj._ref)

    def __iter__(self):
        """ Do not try to add elements during iteration """
        for elem in self._elements_by_ref.values():
            # Skip the Undefined entry
            if elem is not None:
                yield elem

    def __getitem__(self, ref):
        return self._elements_by_ref[ref]

    def __contains__(self, item):
        return item._key in self._elements_by_key

    def __len__(self):
        return len(self._elements_by_ref) - 1


class _StringRegistry(_RefRegistry):
    """
    This is a special registry for strings.
    """
    def __init__(self, definition_registry):
        super(_StringRegistry, self).__init__(definition_registry, Undefined.STRING,
                                              definitions.String)

    def __getitem__(self, ref):
        if self._elements_by_ref[ref] is None:
            return None
        return self._elements_by_ref[ref].name

    def get_ref(self, name, writer=None):
        """
        Get a definition object, and write it in the writer if it was not yet created.
        This means, that all previously created definitions should already be written to the writer.
        Returns the ref of the definition for simplicity.
        """
        if not isinstance(name, str):
            raise TypeError(f"trying to get string reference for non-str type {type(name)}")
        key = definitions.String._new_key(self, name)
        try:
            return self._elements_by_key[key]._ref
        except KeyError:
            pass
        obj = self.create(name, _type=definitions.String)
        if writer:
            obj.write(writer, self._definition_registry)
        return obj._ref


class DefinitionRegistry(object):
    """
        This class stores all definitions.
    """
    def __init__(self):
        self.clock_properties = None
        self._strings = _StringRegistry(self)
        self._metrics = _RefRegistry(self, Undefined.METRIC)
        self._comms = _RefRegistry(self, Undefined.COMM)
        self._io_files = _RefRegistry(self, Undefined.IO_FILE)
        self._io_paradigms = _RefRegistry(self, Undefined.IO_PARADIGM, definitions.IoParadigm)
        self._attributes = _RefRegistry(self, Undefined.ATTRIBUTE, definitions.Attribute)
        self._system_tree_nodes = _RefRegistry(self, Undefined.SYSTEM_TREE_NODE, definitions.SystemTreeNode)
        self._location_groups = _RefRegistry(self, Undefined.LOCATION_GROUP, definitions.LocationGroup)
        self._locations = _RefRegistry(self, Undefined.LOCATION, definitions.Location)
        self._regions = _RefRegistry(self, Undefined.REGION, definitions.Region)
        self._callsites = _RefRegistry(self, Undefined.CALLSITE, definitions.Callsite)
        self._callpaths = _RefRegistry(self, Undefined.CALLPATH, definitions.Callpath)
        self._groups = _RefRegistry(self, Undefined.GROUP, definitions.Group)
        self._metric_members = _RefRegistry(self, Undefined.METRIC_MEMBER, definitions.MetricMember)
        self._parameters = _RefRegistry(self, Undefined.PARAMETER, definitions.Parameter)
        self._rma_wins = _RefRegistry(self, Undefined.RMA_WIN, definitions.RmaWin)
        self._cart_dimensions = _RefRegistry(self, Undefined.CART_DIMENSION, definitions.CartDimension)
        self._cart_topologies = _RefRegistry(self, Undefined.CART_TOPOLOGY, definitions.CartTopology)
        self._source_code_locations = _RefRegistry(self, Undefined.SOURCE_CODE_LOCATION, definitions.SourceCodeLocation)
        self._calling_contexts = _RefRegistry(self, Undefined.CALLING_CONTEXT, definitions.CallingContext)
        self._interrupt_generators = _RefRegistry(self, Undefined.INTERRUPT_GENERATOR, definitions.InterruptGenerator)
        self._io_handles = _RefRegistry(self, Undefined.IO_HANDLE, definitions.IoHandle)

        self._paradigms = _ListRegistry(self, definitions.Paradigm)
        self._paradigm_properties = _ListRegistry(self, definitions.ParadigmProperty)
        self._metric_class_recorders = _ListRegistry(self, definitions.MetricClassRecorder)
        self._system_tree_node_properties = _ListRegistry(self, definitions.SystemTreeNodeProperty)
        self._system_tree_node_domains = _ListRegistry(self, definitions.SystemTreeNodeDomain)
        self._location_group_properties = _ListRegistry(self, definitions.LocationGroupProperty)
        self._location_properties = _ListRegistry(self, definitions.LocationProperty)
        self._cart_coordinates = _ListRegistry(self, definitions.CartCoordinate)
        self._calling_context_properties = _ListRegistry(self, definitions.CallingContextProperty)
        self._io_file_properties = _ListRegistry(self, definitions.IoFileProperty)
        self._io_pre_created_handle_states = _ListRegistry(self, definitions.IoPreCreatedHandleState)
        self._callpath_parameters = _ListRegistry(self, definitions.CallpathParameter)

    def write(self, writer):
        # Warning, the order here is important.
        # That is sometimes not trivial: Groups can refer metrics, but metric instances can
        # refer to groups... So we first write the non-metric groups and then, after metrics,
        # the metric groups
        # Note: The split writing metric groups may violate the strong monotonic order of
        # definition reference

        # Note: We do not write metric classes and instances separately. The order given by the
        # reference will always reflect dependencies from instances on classes
        for thing in chain(self._strings,
                           self._paradigms,
                           self._attributes,
                           self._system_tree_nodes,
                           self._system_tree_node_properties,
                           self._system_tree_node_domains,
                           self._location_groups,
                           self._location_group_properties,
                           self._locations,
                           self._location_properties,
                           self._regions,
                           filter(lambda g: g.group_type != GroupType.METRIC, self._groups),
                           self._metric_members,
                           self._metrics,
                           self._metric_class_recorders,
                           filter(lambda g: g.group_type == GroupType.METRIC, self._groups),
                           self._comms,
                           self._parameters,
                           self._callsites,
                           self._callpaths,
                           self._callpath_parameters,
                           self._rma_wins,
                           self._cart_dimensions,
                           self._cart_topologies,
                           self._cart_coordinates,
                           self._source_code_locations,
                           self._calling_contexts,
                           self._calling_context_properties,
                           self._interrupt_generators,
                           self._paradigm_properties,
                           self._io_paradigms,
                           self._io_files,
                           self._io_file_properties,
                           self._io_handles,
                           self._io_pre_created_handle_states):
            thing.write(writer, self)
        # Make sure all location writers are created, otherwise files will be missing
        for location in self._locations:
            writer._archive.event_writer_from_location(location)

    def _set_global_def_reader_callbacks(self, cbs):
        _otf2.GlobalDefReaderCallbacks_SetClockPropertiesCallback(
            cbs, self._set_clock_properties)
        _otf2.GlobalDefReaderCallbacks_SetMetricClassCallback(
            cbs, self._metrics._callback_gen(definitions.MetricClass))
        _otf2.GlobalDefReaderCallbacks_SetMetricInstanceCallback(
            cbs, self._metrics._callback_gen(definitions.MetricInstance))
        _otf2.GlobalDefReaderCallbacks_SetCommCallback(
            cbs, self._comms._callback_gen(definitions.Comm))
        _otf2.GlobalDefReaderCallbacks_SetIoRegularFileCallback(
            cbs, self._io_files._callback_gen(definitions.IoRegularFile))
        _otf2.GlobalDefReaderCallbacks_SetIoDirectoryCallback(
            cbs, self._io_files._callback_gen(definitions.IoDirectory))
        _otf2.GlobalDefReaderCallbacks_SetInterCommCallback(
            cbs, self._comms._callback_gen(definitions.InterComm))
        _otf2.GlobalDefReaderCallbacks_SetParadigmCallback(
            cbs, self._paradigms._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetParadigmPropertyCallback(
            cbs, self._paradigm_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetIoParadigmCallback(
            cbs, self._io_paradigms._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetStringCallback(
            cbs, self._strings._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetAttributeCallback(
            cbs, self._attributes._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetSystemTreeNodeCallback(
            cbs, self._system_tree_nodes._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetLocationGroupCallback(
            cbs, self._location_groups._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetLocationCallback(
            cbs, self._locations._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetRegionCallback(
            cbs, self._regions._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCallsiteCallback(
            cbs, self._callsites._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCallpathCallback(
            cbs, self._callpaths._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetGroupCallback(
            cbs, self._groups._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetMetricMemberCallback(
            cbs, self._metric_members._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetParameterCallback(
            cbs, self._parameters._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetRmaWinCallback(
            cbs, self._rma_wins._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetMetricClassRecorderCallback(
            cbs, self._metric_class_recorders._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetSystemTreeNodePropertyCallback(
            cbs, self._system_tree_node_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetSystemTreeNodeDomainCallback(
            cbs, self._system_tree_node_domains._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetLocationGroupPropertyCallback(
            cbs, self._location_group_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetLocationPropertyCallback(
            cbs, self._location_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCartDimensionCallback(
            cbs, self._cart_dimensions._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCartTopologyCallback(
            cbs, self._cart_topologies._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCartCoordinateCallback(
            cbs, self._cart_coordinates._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetSourceCodeLocationCallback(
            cbs, self._source_code_locations._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCallingContextCallback(
            cbs, self._calling_contexts._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCallingContextPropertyCallback(
            cbs, self._calling_context_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetInterruptGeneratorCallback(
            cbs, self._interrupt_generators._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetIoFilePropertyCallback(
            cbs, self._io_file_properties._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetIoHandleCallback(
            cbs, self._io_handles._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetIoPreCreatedHandleStateCallback(
            cbs, self._io_pre_created_handle_states._callback_gen())
        _otf2.GlobalDefReaderCallbacks_SetCallpathParameterCallback(
            cbs, self._callpath_parameters._callback_gen())

    def _set_clock_properties(self, _, *args):
        if self.clock_properties is not None:
            raise TraceReaderError("Duplicate ClockProperties entry.")
        self.clock_properties = ClockProperties(None, *args)

    def _string(self, name):
        """
        This function is used internally to intern String definitions.
        They are not exposed publicly, instead, plain str objects are used
        """
        assert type(name) is str
        return self._strings.get(name)

    def _get(self, def_type, ref):
        return self._registry_for_type(def_type)[ref]

    def _registry_for_type(self, def_type):
        if def_type is str:
            return self._strings
        return getattr(self, def_type._registry_name)

    def metric(self, name, *args, **kwargs):
        """
            A convenience method to create a magic metric.
            Details may change in future.

            :param str name: The name for the metric, required argument
        """
        return self.metric_class([self.metric_member(name, *args, **kwargs)])

    @_use_doc_from(definitions.MetricClass)
    def metric_class(self, *args, **kwargs):
        return self._metrics.get(*args, _type=definitions.MetricClass, **kwargs)

    @_use_doc_from(definitions.MetricInstance)
    def metric_instance(self, *args, **kwargs):
        return self._metrics.get(*args, _type=definitions.MetricInstance, **kwargs)

    @_use_doc_from(definitions.Comm)
    def comm(self, *args, **kwargs):
        return self._comms.get(*args, _type=definitions.Comm, **kwargs)

    @_use_doc_from(definitions.IoRegularFile)
    def io_regular_file(self, *args, **kwargs):
        return self._io_files.get(*args, _type=definitions.IoRegularFile, **kwargs)

    @_use_doc_from(definitions.IoDirectory)
    def io_directory(self, *args, **kwargs):
        return self._io_files.get(*args, _type=definitions.IoDirectory, **kwargs)

    @_use_doc_from(definitions.InterComm)
    def inter_comm(self, *args, **kwargs):
        return self._comms.get(*args, _type=definitions.InterComm, **kwargs)

    @_use_doc_from(definitions.IoParadigm)
    def io_paradigm(self, *args, **kwargs):
        return self._io_paradigms.get(*args, **kwargs)

    @_use_doc_from(definitions.Attribute)
    def attribute(self, *args, **kwargs):
        return self._attributes.get(*args, **kwargs)

    @_use_doc_from(definitions.SystemTreeNode)
    def system_tree_node(self, *args, **kwargs):
        return self._system_tree_nodes.get(*args, **kwargs)

    @_use_doc_from(definitions.LocationGroup)
    def location_group(self, *args, **kwargs):
        return self._location_groups.get(*args, **kwargs)

    @_use_doc_from(definitions.Location)
    def location(self, *args, **kwargs):
        return self._locations.get(*args, **kwargs)

    @_use_doc_from(definitions.Region)
    def region(self, *args, **kwargs):
        return self._regions.get(*args, **kwargs)

    @_use_doc_from(definitions.Callsite)
    def callsite(self, *args, **kwargs):
        return self._callsites.get(*args, **kwargs)

    @_use_doc_from(definitions.Callpath)
    def callpath(self, *args, **kwargs):
        return self._callpaths.get(*args, **kwargs)

    @_use_doc_from(definitions.Group)
    def group(self, *args, **kwargs):
        return self._groups.get(*args, **kwargs)

    @_use_doc_from(definitions.MetricMember)
    def metric_member(self, *args, **kwargs):
        return self._metric_members.get(*args, **kwargs)

    @_use_doc_from(definitions.Parameter)
    def parameter(self, *args, **kwargs):
        return self._parameters.get(*args, **kwargs)

    @_use_doc_from(definitions.RmaWin)
    def rma_win(self, *args, **kwargs):
        return self._rma_wins.get(*args, **kwargs)

    @_use_doc_from(definitions.CartDimension)
    def cart_dimension(self, *args, **kwargs):
        return self._cart_dimensions.get(*args, **kwargs)

    @_use_doc_from(definitions.CartTopology)
    def cart_topology(self, *args, **kwargs):
        return self._cart_topologies.get(*args, **kwargs)

    @_use_doc_from(definitions.SourceCodeLocation)
    def source_code_location(self, *args, **kwargs):
        return self._source_code_locations.get(*args, **kwargs)

    @_use_doc_from(definitions.CallingContext)
    def calling_context(self, *args, **kwargs):
        return self._calling_contexts.get(*args, **kwargs)

    @_use_doc_from(definitions.InterruptGenerator)
    def interrupt_generator(self, *args, **kwargs):
        return self._interrupt_generators.get(*args, **kwargs)

    @_use_doc_from(definitions.IoHandle)
    def io_handle(self, *args, **kwargs):
        return self._io_handles.get(*args, **kwargs)

    @_use_doc_from(definitions.Paradigm)
    def paradigm(self, *args, **kwargs):
        return self._paradigms.create(*args, **kwargs)

    @_use_doc_from(definitions.ParadigmProperty)
    def paradigm_property(self, *args, **kwargs):
        return self._paradigm_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.MetricClassRecorder)
    def metric_class_recorder(self, *args, **kwargs):
        return self._metric_class_recorders.create(*args, **kwargs)

    @_use_doc_from(definitions.SystemTreeNodeProperty)
    def system_tree_node_property(self, *args, **kwargs):
        return self._system_tree_node_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.SystemTreeNodeDomain)
    def system_tree_node_domain(self, *args, **kwargs):
        return self._system_tree_node_domains.create(*args, **kwargs)

    @_use_doc_from(definitions.LocationGroupProperty)
    def location_group_property(self, *args, **kwargs):
        return self._location_group_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.LocationProperty)
    def location_property(self, *args, **kwargs):
        return self._location_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.CartCoordinate)
    def cart_coordinate(self, *args, **kwargs):
        return self._cart_coordinates.create(*args, **kwargs)

    @_use_doc_from(definitions.CallingContextProperty)
    def calling_context_property(self, *args, **kwargs):
        return self._calling_context_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.IoFileProperty)
    def io_file_property(self, *args, **kwargs):
        return self._io_file_properties.create(*args, **kwargs)

    @_use_doc_from(definitions.IoPreCreatedHandleState)
    def io_pre_created_handle_state(self, *args, **kwargs):
        return self._io_pre_created_handle_states.create(*args, **kwargs)

    @_use_doc_from(definitions.CallpathParameter)
    def callpath_parameter(self, *args, **kwargs):
        return self._callpath_parameters.create(*args, **kwargs)

    @property
    def strings(self):
        """
            A :py:class:`_Registry` instance representing all String definitons.
        """
        return self._strings

    @property
    def metric_classes(self):
        return filter(lambda m: isinstance(m, definitions.MetricClass), self._metrics)

    @property
    def metric_instances(self):
        return filter(lambda m: isinstance(m, definitions.MetricInstance), self._metrics)

    @property
    def comms(self):
        return filter(lambda m: isinstance(m, definitions.Comm), self._comms)

    @property
    def io_regular_files(self):
        return filter(lambda m: isinstance(m, definitions.IoRegularFile), self._io_files)

    @property
    def io_directories(self):
        return filter(lambda m: isinstance(m, definitions.IoDirectory), self._io_files)

    @property
    def inter_comms(self):
        return filter(lambda m: isinstance(m, definitions.InterComm), self._comms)

    @property
    def io_paradigms(self):
        """
            A :py:class:`_Registry` instance representing all IoParadigm definitons.
        """
        return self._io_paradigms

    @property
    def attributes(self):
        """
            A :py:class:`_Registry` instance representing all Attribute definitons.
        """
        return self._attributes

    @property
    def system_tree_nodes(self):
        """
            A :py:class:`_Registry` instance representing all SystemTreeNode definitons.
        """
        return self._system_tree_nodes

    @property
    def location_groups(self):
        """
            A :py:class:`_Registry` instance representing all LocationGroup definitons.
        """
        return self._location_groups

    @property
    def locations(self):
        """
            A :py:class:`_Registry` instance representing all Location definitons.
        """
        return self._locations

    @property
    def regions(self):
        """
            A :py:class:`_Registry` instance representing all Region definitons.
        """
        return self._regions

    @property
    def callsites(self):
        """
            A :py:class:`_Registry` instance representing all Callsite definitons.
        """
        return self._callsites

    @property
    def callpaths(self):
        """
            A :py:class:`_Registry` instance representing all Callpath definitons.
        """
        return self._callpaths

    @property
    def groups(self):
        """
            A :py:class:`_Registry` instance representing all Group definitons.
        """
        return self._groups

    @property
    def metric_members(self):
        """
            A :py:class:`_Registry` instance representing all MetricMember definitons.
        """
        return self._metric_members

    @property
    def metrics(self):
        """
            A :py:class:`_Registry` instance representing all Metric definitons.
        """
        return self._metrics

    @property
    def comms(self):
        """
            A :py:class:`_Registry` instance representing all Comm definitons.
        """
        return self._comms

    @property
    def parameters(self):
        """
            A :py:class:`_Registry` instance representing all Parameter definitons.
        """
        return self._parameters

    @property
    def rma_wins(self):
        """
            A :py:class:`_Registry` instance representing all RmaWin definitons.
        """
        return self._rma_wins

    @property
    def cart_dimensions(self):
        """
            A :py:class:`_Registry` instance representing all CartDimension definitons.
        """
        return self._cart_dimensions

    @property
    def cart_topologies(self):
        """
            A :py:class:`_Registry` instance representing all CartTopology definitons.
        """
        return self._cart_topologies

    @property
    def source_code_locations(self):
        """
            A :py:class:`_Registry` instance representing all SourceCodeLocation definitons.
        """
        return self._source_code_locations

    @property
    def calling_contexts(self):
        """
            A :py:class:`_Registry` instance representing all CallingContext definitons.
        """
        return self._calling_contexts

    @property
    def interrupt_generators(self):
        """
            A :py:class:`_Registry` instance representing all InterruptGenerator definitons.
        """
        return self._interrupt_generators

    @property
    def io_files(self):
        """
            A :py:class:`_Registry` instance representing all IoFile definitons.
        """
        return self._io_files

    @property
    def io_handles(self):
        """
            A :py:class:`_Registry` instance representing all IoHandle definitons.
        """
        return self._io_handles

    @property
    def paradigms(self):
        """
            A :py:class:`_Registry` instance representing all Paradigm definitons.
        """
        return self._paradigms

    @property
    def paradigm_properties(self):
        """
            A :py:class:`_Registry` instance representing all ParadigmProperty definitons.
        """
        return self._paradigm_properties

    @property
    def metric_class_recorders(self):
        """
            A :py:class:`_Registry` instance representing all MetricClassRecorder definitons.
        """
        return self._metric_class_recorders

    @property
    def system_tree_node_properties(self):
        """
            A :py:class:`_Registry` instance representing all SystemTreeNodeProperty definitons.
        """
        return self._system_tree_node_properties

    @property
    def system_tree_node_domains(self):
        """
            A :py:class:`_Registry` instance representing all SystemTreeNodeDomain definitons.
        """
        return self._system_tree_node_domains

    @property
    def location_group_properties(self):
        """
            A :py:class:`_Registry` instance representing all LocationGroupProperty definitons.
        """
        return self._location_group_properties

    @property
    def location_properties(self):
        """
            A :py:class:`_Registry` instance representing all LocationProperty definitons.
        """
        return self._location_properties

    @property
    def cart_coordinates(self):
        """
            A :py:class:`_Registry` instance representing all CartCoordinate definitons.
        """
        return self._cart_coordinates

    @property
    def calling_context_properties(self):
        """
            A :py:class:`_Registry` instance representing all CallingContextProperty definitons.
        """
        return self._calling_context_properties

    @property
    def io_file_properties(self):
        """
            A :py:class:`_Registry` instance representing all IoFileProperty definitons.
        """
        return self._io_file_properties

    @property
    def io_pre_created_handle_states(self):
        """
            A :py:class:`_Registry` instance representing all IoPreCreatedHandleState definitons.
        """
        return self._io_pre_created_handle_states

    @property
    def callpath_parameters(self):
        """
            A :py:class:`_Registry` instance representing all CallpathParameter definitons.
        """
        return self._callpath_parameters




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import EnumBase, EnumBitset

VERSION_MAJOR = 3
VERSION_MINOR = 1
VERSION_BUGFIX = 1
VERSION_SUFFIX = ""
VERSION = "3.1.1"

# Symbols for undefined values

UNDEFINED_UINT8 = ctypes.c_uint8(-1)
UNDEFINED_INT8 = ctypes.c_int8(0x80)
UNDEFINED_UINT16 = ctypes.c_uint16(-1)
UNDEFINED_INT16 = ctypes.c_int16(0x8000)
UNDEFINED_UINT32 = ctypes.c_uint32(-1)
UNDEFINED_INT32 = ctypes.c_int32(0x80000000)
UNDEFINED_UINT64 = ctypes.c_uint64(-1)
UNDEFINED_INT64 = ctypes.c_int64(0x8000000000000000)

UNDEFINED_TYPE = UNDEFINED_UINT8

# Handle Typedefs

TimeStamp = ctypes.c_uint64
UNDEFINED_TIMESTAMP = UNDEFINED_UINT64

IoParadigmRef = ctypes.c_uint8
UNDEFINED_IO_PARADIGM = IoParadigmRef(UNDEFINED_UINT8.value)

StringRef = ctypes.c_uint32
UNDEFINED_STRING = StringRef(UNDEFINED_UINT32.value)

AttributeRef = ctypes.c_uint32
UNDEFINED_ATTRIBUTE = AttributeRef(UNDEFINED_UINT32.value)

SystemTreeNodeRef = ctypes.c_uint32
UNDEFINED_SYSTEM_TREE_NODE = SystemTreeNodeRef(UNDEFINED_UINT32.value)

LocationGroupRef = ctypes.c_uint32
UNDEFINED_LOCATION_GROUP = LocationGroupRef(UNDEFINED_UINT32.value)

LocationRef = ctypes.c_uint64
UNDEFINED_LOCATION = LocationRef(UNDEFINED_UINT64.value)

RegionRef = ctypes.c_uint32
UNDEFINED_REGION = RegionRef(UNDEFINED_UINT32.value)

CallsiteRef = ctypes.c_uint32
UNDEFINED_CALLSITE = CallsiteRef(UNDEFINED_UINT32.value)

CallpathRef = ctypes.c_uint32
UNDEFINED_CALLPATH = CallpathRef(UNDEFINED_UINT32.value)

GroupRef = ctypes.c_uint32
UNDEFINED_GROUP = GroupRef(UNDEFINED_UINT32.value)

MetricMemberRef = ctypes.c_uint32
UNDEFINED_METRIC_MEMBER = MetricMemberRef(UNDEFINED_UINT32.value)

MetricRef = ctypes.c_uint32
UNDEFINED_METRIC = MetricRef(UNDEFINED_UINT32.value)

CommRef = ctypes.c_uint32
UNDEFINED_COMM = CommRef(UNDEFINED_UINT32.value)

ParameterRef = ctypes.c_uint32
UNDEFINED_PARAMETER = ParameterRef(UNDEFINED_UINT32.value)

RmaWinRef = ctypes.c_uint32
UNDEFINED_RMA_WIN = RmaWinRef(UNDEFINED_UINT32.value)

CartDimensionRef = ctypes.c_uint32
UNDEFINED_CART_DIMENSION = CartDimensionRef(UNDEFINED_UINT32.value)

CartTopologyRef = ctypes.c_uint32
UNDEFINED_CART_TOPOLOGY = CartTopologyRef(UNDEFINED_UINT32.value)

SourceCodeLocationRef = ctypes.c_uint32
UNDEFINED_SOURCE_CODE_LOCATION = SourceCodeLocationRef(UNDEFINED_UINT32.value)

CallingContextRef = ctypes.c_uint32
UNDEFINED_CALLING_CONTEXT = CallingContextRef(UNDEFINED_UINT32.value)

InterruptGeneratorRef = ctypes.c_uint32
UNDEFINED_INTERRUPT_GENERATOR = InterruptGeneratorRef(UNDEFINED_UINT32.value)

IoFileRef = ctypes.c_uint32
UNDEFINED_IO_FILE = IoFileRef(UNDEFINED_UINT32.value)

IoHandleRef = ctypes.c_uint32
UNDEFINED_IO_HANDLE = IoHandleRef(UNDEFINED_UINT32.value)

CHUNK_SIZE_MIN = ctypes.c_uint64(256 * 1024)

CHUNK_SIZE_MAX = ctypes.c_uint64(1024 * 1024 * 16)

class FileMode(EnumBase, ctypes.c_uint8):
    pass

FILEMODE_WRITE = FileMode._construct(0, "WRITE")
FILEMODE_READ = FileMode._construct(1, "READ")
FILEMODE_MODIFY = FileMode._construct(2, "MODIFY")

class Compression(EnumBase, ctypes.c_uint8):
    pass

COMPRESSION_UNDEFINED = Compression._construct(0, "UNDEFINED")
COMPRESSION_NONE = Compression._construct(1, "NONE")
COMPRESSION_ZLIB = Compression._construct(2, "ZLIB")

# CallbackCode is even an enum on the C-Side
# usually used as return value, evtl. should be a restype method?
class CallbackCode(EnumBase, ctypes.c_int):
    pass

CALLBACK_SUCCESS = CallbackCode._construct(0, "SUCCESS")
CALLBACK_INTERRUPT = CallbackCode._construct(1, "INTERRUPT")
CALLBACK_ERROR = CallbackCode._construct(1, "ERROR")

class FlushType(EnumBase, ctypes.c_uint8):
    pass

NO_FLUSH = FlushType._construct(0, "NO_FLUSH")
FLUSH = FlushType._construct(1, "FLUSH")

class Hint(EnumBase, ctypes.c_uint8):
    pass

HINT_GLOBAL_READER = Hint._construct(0, "GLOBAL_READER")

class Boolean(EnumBase, ctypes.c_uint8):
    pass

FALSE = Boolean._construct(0, "FALSE")

class FileType(EnumBase, ctypes.c_uint8):
    pass

FILETYPE_ANCHOR = FileType._construct(0, "ANCHOR")
FILETYPE_GLOBAL_DEFS = FileType._construct(1, "GLOBAL_DEFS")
FILETYPE_LOCAL_DEFS = FileType._construct(2, "LOCAL_DEFS")
FILETYPE_EVENTS = FileType._construct(3, "EVENTS")
FILETYPE_SNAPSHOTS = FileType._construct(4, "SNAPSHOTS")
FILETYPE_THUMBNAIL = FileType._construct(5, "THUMBNAIL")
FILETYPE_MARKER = FileType._construct(6, "MARKER")
FILETYPE_SIONRANKMAP = FileType._construct(7, "SIONRANKMAP")

class FileSubstrate(EnumBase, ctypes.c_uint8):
    pass

SUBSTRATE_UNDEFINED = FileSubstrate._construct(0, "UNDEFINED")
SUBSTRATE_POSIX = FileSubstrate._construct(1, "POSIX")
SUBSTRATE_SION = FileSubstrate._construct(2, "SION")
SUBSTRATE_NONE = FileSubstrate._construct(3, "NONE")

class MappingType(EnumBase, ctypes.c_uint8):
    pass

MAPPING_STRING = MappingType._construct(0, "STRING")
MAPPING_ATTRIBUTE = MappingType._construct(1, "ATTRIBUTE")
MAPPING_LOCATION = MappingType._construct(2, "LOCATION")
MAPPING_REGION = MappingType._construct(3, "REGION")
MAPPING_GROUP = MappingType._construct(4, "GROUP")
MAPPING_METRIC = MappingType._construct(5, "METRIC")
MAPPING_COMM = MappingType._construct(6, "COMM")
MAPPING_PARAMETER = MappingType._construct(7, "PARAMETER")
MAPPING_RMA_WIN = MappingType._construct(8, "RMA_WIN")
MAPPING_SOURCE_CODE_LOCATION = MappingType._construct(9, "SOURCE_CODE_LOCATION")
MAPPING_CALLING_CONTEXT = MappingType._construct(10, "CALLING_CONTEXT")
MAPPING_INTERRUPT_GENERATOR = MappingType._construct(11, "INTERRUPT_GENERATOR")
MAPPING_IO_FILE = MappingType._construct(12, "IO_FILE")
MAPPING_IO_HANDLE = MappingType._construct(13, "IO_HANDLE")
MAPPING_LOCATION_GROUP = MappingType._construct(14, "LOCATION_GROUP")
MAPPING_MAX = MappingType(15)

class Type(EnumBase, ctypes.c_uint8):
    pass

TYPE_NONE = Type._construct(0, "NONE")
TYPE_UINT8 = Type._construct(1, "UINT8")
TYPE_UINT16 = Type._construct(2, "UINT16")
TYPE_UINT32 = Type._construct(3, "UINT32")
TYPE_UINT64 = Type._construct(4, "UINT64")
TYPE_INT8 = Type._construct(5, "INT8")
TYPE_INT16 = Type._construct(6, "INT16")
TYPE_INT32 = Type._construct(7, "INT32")
TYPE_INT64 = Type._construct(8, "INT64")
TYPE_FLOAT = Type._construct(9, "FLOAT")
TYPE_DOUBLE = Type._construct(10, "DOUBLE")
TYPE_STRING = Type._construct(11, "STRING")
TYPE_ATTRIBUTE = Type._construct(12, "ATTRIBUTE")
TYPE_LOCATION = Type._construct(13, "LOCATION")
TYPE_REGION = Type._construct(14, "REGION")
TYPE_GROUP = Type._construct(15, "GROUP")
TYPE_METRIC = Type._construct(16, "METRIC")
TYPE_COMM = Type._construct(17, "COMM")
TYPE_PARAMETER = Type._construct(18, "PARAMETER")
TYPE_RMA_WIN = Type._construct(19, "RMA_WIN")
TYPE_SOURCE_CODE_LOCATION = Type._construct(20, "SOURCE_CODE_LOCATION")
TYPE_CALLING_CONTEXT = Type._construct(21, "CALLING_CONTEXT")
TYPE_INTERRUPT_GENERATOR = Type._construct(22, "INTERRUPT_GENERATOR")
TYPE_IO_FILE = Type._construct(23, "IO_FILE")
TYPE_IO_HANDLE = Type._construct(24, "IO_HANDLE")
TYPE_LOCATION_GROUP = Type._construct(25, "LOCATION_GROUP")

class Paradigm(EnumBase, ctypes.c_uint8):
    pass

PARADIGM_UNKNOWN = Paradigm._construct(0, "UNKNOWN")
PARADIGM_USER = Paradigm._construct(1, "USER")
PARADIGM_COMPILER = Paradigm._construct(2, "COMPILER")
PARADIGM_OPENMP = Paradigm._construct(3, "OPENMP")
PARADIGM_MPI = Paradigm._construct(4, "MPI")
PARADIGM_CUDA = Paradigm._construct(5, "CUDA")
PARADIGM_MEASUREMENT_SYSTEM = Paradigm._construct(6, "MEASUREMENT_SYSTEM")
PARADIGM_PTHREAD = Paradigm._construct(7, "PTHREAD")
PARADIGM_HMPP = Paradigm._construct(8, "HMPP")
PARADIGM_OMPSS = Paradigm._construct(9, "OMPSS")
PARADIGM_HARDWARE = Paradigm._construct(10, "HARDWARE")
PARADIGM_GASPI = Paradigm._construct(11, "GASPI")
PARADIGM_UPC = Paradigm._construct(12, "UPC")
PARADIGM_SHMEM = Paradigm._construct(13, "SHMEM")
PARADIGM_WINTHREAD = Paradigm._construct(14, "WINTHREAD")
PARADIGM_QTTHREAD = Paradigm._construct(15, "QTTHREAD")
PARADIGM_ACETHREAD = Paradigm._construct(16, "ACETHREAD")
PARADIGM_TBBTHREAD = Paradigm._construct(17, "TBBTHREAD")
PARADIGM_OPENACC = Paradigm._construct(18, "OPENACC")
PARADIGM_OPENCL = Paradigm._construct(19, "OPENCL")
PARADIGM_MTAPI = Paradigm._construct(20, "MTAPI")
PARADIGM_SAMPLING = Paradigm._construct(21, "SAMPLING")
PARADIGM_NONE = Paradigm._construct(22, "NONE")
PARADIGM_HIP = Paradigm._construct(23, "HIP")
PARADIGM_KOKKOS = Paradigm._construct(24, "KOKKOS")
PARADIGM_OPENMP_TARGET = Paradigm._construct(25, "OPENMP_TARGET")

class ParadigmClass(EnumBase, ctypes.c_uint8):
    pass

PARADIGM_CLASS_PROCESS = ParadigmClass._construct(0, "PROCESS")
PARADIGM_CLASS_THREAD_FORK_JOIN = ParadigmClass._construct(1, "THREAD_FORK_JOIN")
PARADIGM_CLASS_THREAD_CREATE_WAIT = ParadigmClass._construct(2, "THREAD_CREATE_WAIT")
PARADIGM_CLASS_ACCELERATOR = ParadigmClass._construct(3, "ACCELERATOR")

class ParadigmProperty(EnumBase, ctypes.c_uint8):
    pass

PARADIGM_PROPERTY_COMM_NAME_TEMPLATE = ParadigmProperty._construct(0, "COMM_NAME_TEMPLATE")
PARADIGM_PROPERTY_RMA_WIN_NAME_TEMPLATE = ParadigmProperty._construct(1, "RMA_WIN_NAME_TEMPLATE")
PARADIGM_PROPERTY_RMA_ONLY = ParadigmProperty._construct(2, "RMA_ONLY")

class ThumbnailType(EnumBase, ctypes.c_uint8):
    pass

THUMBNAIL_TYPE_REGION = ThumbnailType._construct(0, "REGION")
THUMBNAIL_TYPE_METRIC = ThumbnailType._construct(1, "METRIC")
THUMBNAIL_TYPE_ATTRIBUTES = ThumbnailType._construct(2, "ATTRIBUTES")
THUMBNAIL_TYPE_COMM = ThumbnailType._construct(3, "COMM")
THUMBNAIL_TYPE_RMA_WIN = ThumbnailType._construct(4, "RMA_WIN")
THUMBNAIL_TYPE_IO_HANDLE = ThumbnailType._construct(5, "IO_HANDLE")


TRUE = Boolean._construct(1, "TRUE")

__all__ = [
    'VERSION_MAJOR',
    'VERSION_MINOR',
    'VERSION_BUGFIX',
    'VERSION_SUFFIX',
    'VERSION',
    'UNDEFINED_UINT8',
    'UNDEFINED_INT8',
    'UNDEFINED_UINT16',
    'UNDEFINED_INT16',
    'UNDEFINED_UINT32',
    'UNDEFINED_INT32',
    'UNDEFINED_UINT64',
    'UNDEFINED_INT64',
    'UNDEFINED_TYPE',
    'TimeStamp',
    'UNDEFINED_TIMESTAMP',
    'IoParadigmRef',
    'UNDEFINED_IO_PARADIGM',
    'StringRef',
    'UNDEFINED_STRING',
    'AttributeRef',
    'UNDEFINED_ATTRIBUTE',
    'SystemTreeNodeRef',
    'UNDEFINED_SYSTEM_TREE_NODE',
    'LocationGroupRef',
    'UNDEFINED_LOCATION_GROUP',
    'LocationRef',
    'UNDEFINED_LOCATION',
    'RegionRef',
    'UNDEFINED_REGION',
    'CallsiteRef',
    'UNDEFINED_CALLSITE',
    'CallpathRef',
    'UNDEFINED_CALLPATH',
    'GroupRef',
    'UNDEFINED_GROUP',
    'MetricMemberRef',
    'UNDEFINED_METRIC_MEMBER',
    'MetricRef',
    'UNDEFINED_METRIC',
    'CommRef',
    'UNDEFINED_COMM',
    'ParameterRef',
    'UNDEFINED_PARAMETER',
    'RmaWinRef',
    'UNDEFINED_RMA_WIN',
    'CartDimensionRef',
    'UNDEFINED_CART_DIMENSION',
    'CartTopologyRef',
    'UNDEFINED_CART_TOPOLOGY',
    'SourceCodeLocationRef',
    'UNDEFINED_SOURCE_CODE_LOCATION',
    'CallingContextRef',
    'UNDEFINED_CALLING_CONTEXT',
    'InterruptGeneratorRef',
    'UNDEFINED_INTERRUPT_GENERATOR',
    'IoFileRef',
    'UNDEFINED_IO_FILE',
    'IoHandleRef',
    'UNDEFINED_IO_HANDLE',
    'CHUNK_SIZE_MIN',
    'CHUNK_SIZE_MAX',
    'FileMode',
    'FILEMODE_WRITE',
    'FILEMODE_READ',
    'FILEMODE_MODIFY',
    'Compression',
    'COMPRESSION_UNDEFINED',
    'COMPRESSION_NONE',
    'COMPRESSION_ZLIB',
    'CallbackCode',
    'CALLBACK_SUCCESS',
    'CALLBACK_INTERRUPT',
    'CALLBACK_ERROR',
    'FlushType',
    'NO_FLUSH',
    'FLUSH',
    'Boolean',
    'FALSE',
    'TRUE',
    'FileType',
    'FILETYPE_ANCHOR',
    'FILETYPE_GLOBAL_DEFS',
    'FILETYPE_LOCAL_DEFS',
    'FILETYPE_EVENTS',
    'FILETYPE_SNAPSHOTS',
    'FILETYPE_THUMBNAIL',
    'FILETYPE_MARKER',
    'FILETYPE_SIONRANKMAP',
    'FileSubstrate',
    'SUBSTRATE_UNDEFINED',
    'SUBSTRATE_POSIX',
    'SUBSTRATE_SION',
    'SUBSTRATE_NONE',
    'MappingType',
    'MAPPING_STRING',
    'MAPPING_ATTRIBUTE',
    'MAPPING_LOCATION',
    'MAPPING_REGION',
    'MAPPING_GROUP',
    'MAPPING_METRIC',
    'MAPPING_COMM',
    'MAPPING_PARAMETER',
    'MAPPING_RMA_WIN',
    'MAPPING_SOURCE_CODE_LOCATION',
    'MAPPING_CALLING_CONTEXT',
    'MAPPING_INTERRUPT_GENERATOR',
    'MAPPING_IO_FILE',
    'MAPPING_IO_HANDLE',
    'MAPPING_LOCATION_GROUP',
    'MAPPING_MAX',
    'Type',
    'TYPE_NONE',
    'TYPE_UINT8',
    'TYPE_UINT16',
    'TYPE_UINT32',
    'TYPE_UINT64',
    'TYPE_INT8',
    'TYPE_INT16',
    'TYPE_INT32',
    'TYPE_INT64',
    'TYPE_FLOAT',
    'TYPE_DOUBLE',
    'TYPE_STRING',
    'TYPE_ATTRIBUTE',
    'TYPE_LOCATION',
    'TYPE_REGION',
    'TYPE_GROUP',
    'TYPE_METRIC',
    'TYPE_COMM',
    'TYPE_PARAMETER',
    'TYPE_RMA_WIN',
    'TYPE_SOURCE_CODE_LOCATION',
    'TYPE_CALLING_CONTEXT',
    'TYPE_INTERRUPT_GENERATOR',
    'TYPE_IO_FILE',
    'TYPE_IO_HANDLE',
    'TYPE_LOCATION_GROUP',
    'Paradigm',
    'PARADIGM_UNKNOWN',
    'PARADIGM_USER',
    'PARADIGM_COMPILER',
    'PARADIGM_OPENMP',
    'PARADIGM_MPI',
    'PARADIGM_CUDA',
    'PARADIGM_MEASUREMENT_SYSTEM',
    'PARADIGM_PTHREAD',
    'PARADIGM_HMPP',
    'PARADIGM_OMPSS',
    'PARADIGM_HARDWARE',
    'PARADIGM_GASPI',
    'PARADIGM_UPC',
    'PARADIGM_SHMEM',
    'PARADIGM_WINTHREAD',
    'PARADIGM_QTTHREAD',
    'PARADIGM_ACETHREAD',
    'PARADIGM_TBBTHREAD',
    'PARADIGM_OPENACC',
    'PARADIGM_OPENCL',
    'PARADIGM_MTAPI',
    'PARADIGM_SAMPLING',
    'PARADIGM_NONE',
    'PARADIGM_HIP',
    'PARADIGM_KOKKOS',
    'PARADIGM_OPENMP_TARGET',
    'ParadigmClass',
    'PARADIGM_CLASS_PROCESS',
    'PARADIGM_CLASS_THREAD_FORK_JOIN',
    'PARADIGM_CLASS_THREAD_CREATE_WAIT',
    'PARADIGM_CLASS_ACCELERATOR',
    'ParadigmProperty',
    'PARADIGM_PROPERTY_COMM_NAME_TEMPLATE',
    'PARADIGM_PROPERTY_RMA_WIN_NAME_TEMPLATE',
    'PARADIGM_PROPERTY_RMA_ONLY',
    'ThumbnailType',
    'THUMBNAIL_TYPE_REGION',
    'THUMBNAIL_TYPE_METRIC',
    'THUMBNAIL_TYPE_ATTRIBUTES',
    'THUMBNAIL_TYPE_COMM',
    'THUMBNAIL_TYPE_RMA_WIN',
    'THUMBNAIL_TYPE_IO_HANDLE',
    'Hint',
    'HINT_GLOBAL_READER'
]




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .AttributeList import AttributeList


class EventSizeEstimator(ctypes.Structure):
    pass

def EventSizeEstimator_New():
    c_New = conf.lib.OTF2_EventSizeEstimator_New
    c_New.argtypes = []
    c_New.restype = ctypes.POINTER(EventSizeEstimator)
    # NOTE: Do not errcheck here. This function returns a value, not an error code
    return c_EventSizeEstimator_New()

def EventSizeEstimator_Delete(estimator):
    c_Delete = conf.lib.OTF2_EventSizeEstimator_Delete
    c_Delete.argtypes = [ ctypes.POINTER(EventSizeEstimator) ]
    c_Delete.restype = ErrorCode
    c_Delete.errcheck = HandleErrorCode
    c_Delete(estimator)

@otf2 for entry in MappingType.entries:
def EventSizeEstimator_SetNumberOf@@entry.def.ref_name@@Definitions(estimator, numberOf@@entry.def.ref_name@@Definitions):
    c_SetNumberOf@@entry.def.ref_name@@Definitions = conf.lib.OTF2_EventSizeEstimator_SetNumberOf@@entry.def.ref_name@@Definitions
    c_SetNumberOf@@entry.def.ref_name@@Definitions.argtypes = [ ctypes.POINTER(EventSizeEstimator), @@entry.def.self_attr.type.py_argtype@@ ]
    c_SetNumberOf@@entry.def.ref_name@@Definitions.restype = ErrorCode
    c_SetNumberOf@@entry.def.ref_name@@Definitions.errcheck = HandleErrorCode
    c_SetNumberOf@@entry.def.ref_name@@Definitions(estimator, numberOf@@entry.def.ref_name@@Definitions)

@otf2 endfor
def EventSizeEstimator_GetDefChunkSize(estimator):
    c_GetDefChunkSize = conf.lib.OTF2_EventSizeEstimator_GetDefChunkSize
    c_GetDefChunkSize.argtypes = [ ctypes.POINTER(EventSizeEstimator) ]
    c_GetDefChunkSize.restype = ctypes.c_size_t
    return c_GetDefChunkSize(estimator)

def EventSizeEstimator_GetSizeOfTimestamp(estimator):
    c_GetSizeOfTimestamp = conf.lib.OTF2_EventSizeEstimator_GetSizeOfTimestamp
    c_GetSizeOfTimestamp.argtypes = [ ctypes.POINTER(EventSizeEstimator) ]
    c_GetSizeOfTimestamp.restype = ctypes.c_size_t
    return c_EventSizeEstimator_GetSizeOfTimestamp(estimator)

def EventSizeEstimator_GetSizeOfAttributeList(estimator, attributeList):
    c_GetSizeOfAttributeList = conf.lib.OTF2_EventSizeEstimator_GetSizeOfAttributeList
    c_GetSizeOfAttributeList.argtypes = [ ctypes.POINTER(EventSizeEstimator), ctypes.POINTER(AttributeList) ]
    c_GetSizeOfAttributeList.restype = ctypes.c_size_t
    return c_GetSizeOfAttributeList(estimator, attributeList)

@otf2 for event in events:
def EventSizeEstimator_GetSizeOf@@event.name@@Event(estimator@@event.py_funcargs(attributes=event.array_attributes, resolve_array_attributes=False)@@):
    c_GetSizeOf@@event.name@@Event = conf.lib.OTF2_EventSizeEstimator_GetSizeOf@@event.name@@Event
    c_GetSizeOf@@event.name@@Event.argtypes = [ ctypes.POINTER(EventSizeEstimator)@@event.py_funcargtypes(attributes=event.array_attributes, resolve_array_attributes=False)@@ ]
    c_GetSizeOf@@event.name@@Event.restype = ctypes.c_size_t
    return c_EventSizeEstimator_GetSizeOf@@event.name@@Event(estimator@@event.py_callargs(attributes=event.array_attributes, resolve_array_attributes=False)@@)

@otf2 endfor
__all__ = [
    'EventSizeEstimator',
    'EventSizeEstimator_New',
    'EventSizeEstimator_Delete',
    'EventSizeEstimator_GetDefChunkSize',
    'EventSizeEstimator_GetSizeOfTimestamp',
    'EventSizeEstimator_GetSizeOfAttributeList',
    @otf2 for entry in MappingType.entries:
    'EventSizeEstimator_SetNumberOf@@entry.def.ref_name@@Definitions',
    @otf2 endfor
    @otf2 for event in events:
    'EventSizeEstimator_GetSizeOf@@event.name@@Event',
    @otf2 endfor
]



#!/usr/bin/env python
#
# This file is part of the Score-P software (http://www.score-p.org)
#
# Copyright (c) 2015-2017,
# Technische Universitaet Dresden, Germany
#
# This software may be modified and distributed under the terms of
# a BSD-style license.  See the COPYING file in the package base
# directory for details.
#

import otf2


with otf2.reader.open('TestArchive/traces.otf2') as trace:
    print("Read {} string definitions".format(len(trace.definitions.strings)))

    for location, event in trace.events:
        print("Encountered {} event {} on {}".format(type(event).__name__,
                                                              event, location))




# Generated by OTF2 Template Engine

import _otf2

from .enums import Type


def parse(registry, union_value, type_id):
@otf2 for entry in Type.entries if entry.type:
    if type_id == Type.@@entry.suffix@@:
        return union_value.@@entry.type.lower@@
@otf2 endfor
@otf2 for entry in MappingType.entries:
    if type_id == Type.@@entry.def.ref_upper@@:
        return registry.@@entry.def.ref_lower@@s[union_value.@@entry.def.ref_lname@@Ref]
@otf2 endfor
    raise TypeError("Unexpected type.")


def make(registry, value, type_id, writer=None):
    """
    :param registry:
    :param value:
    :param otf2.Type type_id:
    :param writer: If set, the string value will be written. Otherwise just added to the registry
    :return:
    """
    # values are expected to be int, float, or a Definition subclass
    union = _otf2.AttributeValue()
    if type_id == Type.UINT8:
        union.uint8 = int(value)
    elif type_id == Type.UINT16:
        union.uint16 = int(value)
    elif type_id == Type.UINT32:
        union.uint32 = int(value)
    elif type_id == Type.UINT64:
        union.uint64 = int(value)
    elif type_id == Type.INT8:
        union.int8 = int(value)
    elif type_id == Type.INT16:
        union.int16 = int(value)
    elif type_id == Type.INT32:
        union.int32 = int(value)
    elif type_id == Type.INT64:
        union.int64 = int(value)
    elif type_id == Type.FLOAT:
        union.float32 = float(value)
    elif type_id == Type.DOUBLE:
        union.float64 = float(value)
    elif type_id == Type.STRING:
        union.stringRef = registry.strings.get_ref(value, writer)
@otf2 for entry in MappingType.entries if entry.name != 'OTF2_MAPPING_STRING':
    elif type_id == Type.@@entry.suffix@@:
        union.@@entry.def.ref_lname@@Ref = value._ref
@otf2 endfor
    else:
        raise TypeError("Unexpected type id given")

    return union


class AttributeValue(object):
    def __init__(self, value, type_id):
        self.value = value
        self.type_id = type_id

    def as_otf2(self, registry, writer=None):
        return make(registry, self.value, self.type_id, writer)

    def set(self, value, type_id=None):
        # TODO Ideally we would want to check if the value is compatible with the type,
        # but that would introduce a circular dependency to the definitions
        self.value = value
        if type_id is not None:
            self.type_id = type_id

    @classmethod
    def _construct(cls, registry, union_value, type_id):
        return cls(parse(registry, union_value, type_id), type_id)




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .Definitions import *
from .AttributeValue import AttributeValue


class GlobalDefWriter(ctypes.Structure):
    pass

def GlobalDefWriter_GetNumberOfDefinitions(writerHandle, numberOfDefinitions = ctypes.c_uint64()):
    c_GetNumberOfDefinitions = conf.lib.OTF2_GlobalDefWriter_GetNumberOfDefinitions
    c_GetNumberOfDefinitions.argtypes = [ ctypes.POINTER(GlobalDefWriter), ctypes.POINTER(ctypes.c_uint64) ]
    c_GetNumberOfDefinitions.restype = ErrorCode
    c_GetNumberOfDefinitions.errcheck = HandleErrorCode
    c_GetNumberOfDefinitions(writerHandle, ctypes.byref(numberOfDefinitions))
    return numberOfDefinitions.value

def GlobalDefWriter_GetNumberOfLocations(writerHandle, numberOfLocations = ctypes.c_uint64()):
    c_GetNumberOfLocations = conf.lib.OTF2_GlobalDefWriter_GetNumberOfLocations
    c_GetNumberOfLocations.argtypes = [ ctypes.POINTER(GlobalDefWriter), ctypes.POINTER(ctypes.c_uint64) ]
    c_GetNumberOfLocations.restype = ErrorCode
    c_GetNumberOfLocations.errcheck = HandleErrorCode
    c_GetNumberOfLocations(writerHandle, ctypes.byref(numberOfLocations))
    return numberOfLocations.value

def GlobalDefWriter_WriteClockProperties(writerHandle, timerResolution, globalOffset, traceLength, realtimeTimestamp):
    c_WriteClockProperties = conf.lib.OTF2_GlobalDefWriter_WriteClockProperties
    c_WriteClockProperties.argtypes = [ ctypes.POINTER(GlobalDefWriter), ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64 ]
    c_WriteClockProperties.restype = ErrorCode
    c_WriteClockProperties.errcheck = HandleErrorCode
    c_WriteClockProperties(writerHandle, timerResolution, globalOffset, traceLength, realtimeTimestamp)

def GlobalDefWriter_WriteMappingTable(writerHandle, mappingType, idMap):
    c_WriteMappingTable = conf.lib.OTF2_GlobalDefWriter_WriteMappingTable
    c_WriteMappingTable.argtypes = [ ctypes.POINTER(GlobalDefWriter), MappingType, ctypes.POINTER(IdMap) ]
    c_WriteMappingTable.restype = ErrorCode
    c_WriteMappingTable.errcheck = HandleErrorCode
    c_WriteMappingTable(writerHandle, mappingType, idMap)

def GlobalDefWriter_WriteClockOffset(writerHandle, time, offset, standardDeviation):
    c_WriteClockOffset = conf.lib.OTF2_GlobalDefWriter_WriteClockOffset
    c_WriteClockOffset.argtypes = [ ctypes.POINTER(GlobalDefWriter), TimeStamp, ctypes.c_int64, ctypes.c_double ]
    c_WriteClockOffset.restype = ErrorCode
    c_WriteClockOffset.errcheck = HandleErrorCode
    c_WriteClockOffset(writerHandle, time, offset, standardDeviation)

def GlobalDefWriter_WriteParadigm(writerHandle, paradigm, name, paradigmClass):
    c_WriteParadigm = conf.lib.OTF2_GlobalDefWriter_WriteParadigm
    c_WriteParadigm.argtypes = [ ctypes.POINTER(GlobalDefWriter), Paradigm, StringRef, ParadigmClass ]
    c_WriteParadigm.restype = ErrorCode
    c_WriteParadigm.errcheck = HandleErrorCode
    c_WriteParadigm(writerHandle, paradigm, name, paradigmClass)

def GlobalDefWriter_WriteParadigmProperty(writerHandle, paradigm, property, type, value):
    c_WriteParadigmProperty = conf.lib.OTF2_GlobalDefWriter_WriteParadigmProperty
    c_WriteParadigmProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), Paradigm, ParadigmProperty, Type, AttributeValue._arg_type ]
    c_WriteParadigmProperty.restype = ErrorCode
    c_WriteParadigmProperty.errcheck = HandleErrorCode
    c_WriteParadigmProperty(writerHandle, paradigm, property, type, value._arg_value)

def GlobalDefWriter_WriteIoParadigm(writerHandle, self, identification, name, ioParadigmClass, ioParadigmFlags, properties, types, values):
    numberOfProperties = len(properties)
    assert numberOfProperties == len(types) == len(values)
    properties_array = (IoParadigmProperty * numberOfProperties)()
    properties_array[:] = properties
    types_array = (Type * numberOfProperties)()
    types_array[:] = types
    values_array = (AttributeValue * numberOfProperties)()
    values_array[:] = values

    c_WriteIoParadigm = conf.lib.OTF2_GlobalDefWriter_WriteIoParadigm
    c_WriteIoParadigm.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoParadigmRef, StringRef, StringRef, IoParadigmClass, IoParadigmFlag, ctypes.c_uint8, ctypes.POINTER(IoParadigmProperty), ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_WriteIoParadigm.restype = ErrorCode
    c_WriteIoParadigm.errcheck = HandleErrorCode
    c_WriteIoParadigm(writerHandle, self, identification, name, ioParadigmClass, ioParadigmFlags,
                      numberOfProperties, properties_array, types_array, values_array)

def GlobalDefWriter_WriteString(writerHandle, self, string):
    c_WriteString = conf.lib.OTF2_GlobalDefWriter_WriteString
    c_WriteString.argtypes = [ ctypes.POINTER(GlobalDefWriter), StringRef, StrParam ]
    c_WriteString.restype = ErrorCode
    c_WriteString.errcheck = HandleErrorCode
    c_WriteString(writerHandle, self, string)

def GlobalDefWriter_WriteAttribute(writerHandle, self, name, description, type):
    c_WriteAttribute = conf.lib.OTF2_GlobalDefWriter_WriteAttribute
    c_WriteAttribute.argtypes = [ ctypes.POINTER(GlobalDefWriter), AttributeRef, StringRef, StringRef, Type ]
    c_WriteAttribute.restype = ErrorCode
    c_WriteAttribute.errcheck = HandleErrorCode
    c_WriteAttribute(writerHandle, self, name, description, type)

def GlobalDefWriter_WriteSystemTreeNode(writerHandle, self, name, className, parent):
    c_WriteSystemTreeNode = conf.lib.OTF2_GlobalDefWriter_WriteSystemTreeNode
    c_WriteSystemTreeNode.argtypes = [ ctypes.POINTER(GlobalDefWriter), SystemTreeNodeRef, StringRef, StringRef, SystemTreeNodeRef ]
    c_WriteSystemTreeNode.restype = ErrorCode
    c_WriteSystemTreeNode.errcheck = HandleErrorCode
    c_WriteSystemTreeNode(writerHandle, self, name, className, parent)

def GlobalDefWriter_WriteLocationGroup(writerHandle, self, name, locationGroupType, systemTreeParent, creatingLocationGroup):
    c_WriteLocationGroup = conf.lib.OTF2_GlobalDefWriter_WriteLocationGroup
    c_WriteLocationGroup.argtypes = [ ctypes.POINTER(GlobalDefWriter), LocationGroupRef, StringRef, LocationGroupType, SystemTreeNodeRef, LocationGroupRef ]
    c_WriteLocationGroup.restype = ErrorCode
    c_WriteLocationGroup.errcheck = HandleErrorCode
    c_WriteLocationGroup(writerHandle, self, name, locationGroupType, systemTreeParent, creatingLocationGroup)

def GlobalDefWriter_WriteLocation(writerHandle, self, name, locationType, numberOfEvents, locationGroup):
    c_WriteLocation = conf.lib.OTF2_GlobalDefWriter_WriteLocation
    c_WriteLocation.argtypes = [ ctypes.POINTER(GlobalDefWriter), LocationRef, StringRef, LocationType, ctypes.c_uint64, LocationGroupRef ]
    c_WriteLocation.restype = ErrorCode
    c_WriteLocation.errcheck = HandleErrorCode
    c_WriteLocation(writerHandle, self, name, locationType, numberOfEvents, locationGroup)

def GlobalDefWriter_WriteRegion(writerHandle, self, name, canonicalName, description, regionRole, paradigm, regionFlags, sourceFile, beginLineNumber, endLineNumber):
    c_WriteRegion = conf.lib.OTF2_GlobalDefWriter_WriteRegion
    c_WriteRegion.argtypes = [ ctypes.POINTER(GlobalDefWriter), RegionRef, StringRef, StringRef, StringRef, RegionRole, Paradigm, RegionFlag, StringRef, ctypes.c_uint32, ctypes.c_uint32 ]
    c_WriteRegion.restype = ErrorCode
    c_WriteRegion.errcheck = HandleErrorCode
    c_WriteRegion(writerHandle, self, name, canonicalName, description, regionRole, paradigm, regionFlags, sourceFile, beginLineNumber, endLineNumber)

def GlobalDefWriter_WriteCallsite(writerHandle, self, sourceFile, lineNumber, enteredRegion, leftRegion):
    c_WriteCallsite = conf.lib.OTF2_GlobalDefWriter_WriteCallsite
    c_WriteCallsite.argtypes = [ ctypes.POINTER(GlobalDefWriter), CallsiteRef, StringRef, ctypes.c_uint32, RegionRef, RegionRef ]
    c_WriteCallsite.restype = ErrorCode
    c_WriteCallsite.errcheck = HandleErrorCode
    c_WriteCallsite(writerHandle, self, sourceFile, lineNumber, enteredRegion, leftRegion)

def GlobalDefWriter_WriteCallpath(writerHandle, self, parent, region):
    c_WriteCallpath = conf.lib.OTF2_GlobalDefWriter_WriteCallpath
    c_WriteCallpath.argtypes = [ ctypes.POINTER(GlobalDefWriter), CallpathRef, CallpathRef, RegionRef ]
    c_WriteCallpath.restype = ErrorCode
    c_WriteCallpath.errcheck = HandleErrorCode
    c_WriteCallpath(writerHandle, self, parent, region)

def GlobalDefWriter_WriteGroup(writerHandle, self, name, groupType, paradigm, groupFlags, members):
    numberOfMembers = len(members)
    array_type = ctypes.c_uint64 * numberOfMembers
    members_array = array_type()
    members_array[:] = members

    c_WriteGroup = conf.lib.OTF2_GlobalDefWriter_WriteGroup
    c_WriteGroup.argtypes = [ ctypes.POINTER(GlobalDefWriter), GroupRef, StringRef, GroupType, Paradigm, GroupFlag,
                              ctypes.c_uint32, array_type ]
    c_WriteGroup.restype = ErrorCode
    c_WriteGroup.errcheck = HandleErrorCode
    c_WriteGroup(writerHandle, self, name, groupType, paradigm, groupFlags, numberOfMembers, members_array)

def GlobalDefWriter_WriteMetricMember(writerHandle, self, name, description, metricType, metricMode, valueType, base, exponent, unit):
    c_WriteMetricMember = conf.lib.OTF2_GlobalDefWriter_WriteMetricMember
    c_WriteMetricMember.argtypes = [ ctypes.POINTER(GlobalDefWriter), MetricMemberRef, StringRef, StringRef, MetricType, MetricMode, Type, Base, ctypes.c_int64, StringRef ]
    c_WriteMetricMember.restype = ErrorCode
    c_WriteMetricMember.errcheck = HandleErrorCode
    c_WriteMetricMember(writerHandle, self, name, description, metricType, metricMode, valueType, base, exponent, unit)

def GlobalDefWriter_WriteMetricClass(writerHandle, self, metricMembers, metricOccurrence, recorderKind):
    numberOfMetrics = len(metricMembers)
    array_type = MetricMemberRef * numberOfMetrics
    metric_member_array = array_type()
    metric_member_array[:] = metricMembers

    c_WriteMetricClass = conf.lib.OTF2_GlobalDefWriter_WriteMetricClass
    c_WriteMetricClass.argtypes = [ ctypes.POINTER(GlobalDefWriter), MetricRef, ctypes.c_uint8, array_type,
                                    MetricOccurrence, RecorderKind ]
    c_WriteMetricClass.restype = ErrorCode
    c_WriteMetricClass.errcheck = HandleErrorCode
    c_WriteMetricClass(writerHandle, MetricRef(self), numberOfMetrics, metric_member_array,
                       metricOccurrence, recorderKind)

def GlobalDefWriter_WriteMetricInstance(writerHandle, self, metricClass, recorder, metricScope, scope):
    c_WriteMetricInstance = conf.lib.OTF2_GlobalDefWriter_WriteMetricInstance
    c_WriteMetricInstance.argtypes = [ ctypes.POINTER(GlobalDefWriter), MetricRef, MetricRef, LocationRef, MetricScope, ctypes.c_uint64 ]
    c_WriteMetricInstance.restype = ErrorCode
    c_WriteMetricInstance.errcheck = HandleErrorCode
    c_WriteMetricInstance(writerHandle, self, metricClass, recorder, metricScope, scope)

def GlobalDefWriter_WriteComm(writerHandle, self, name, group, parent, flags):
    c_WriteComm = conf.lib.OTF2_GlobalDefWriter_WriteComm
    c_WriteComm.argtypes = [ ctypes.POINTER(GlobalDefWriter), CommRef, StringRef, GroupRef, CommRef, CommFlag ]
    c_WriteComm.restype = ErrorCode
    c_WriteComm.errcheck = HandleErrorCode
    c_WriteComm(writerHandle, self, name, group, parent, flags)

def GlobalDefWriter_WriteParameter(writerHandle, self, name, parameterType):
    c_WriteParameter = conf.lib.OTF2_GlobalDefWriter_WriteParameter
    c_WriteParameter.argtypes = [ ctypes.POINTER(GlobalDefWriter), ParameterRef, StringRef, ParameterType ]
    c_WriteParameter.restype = ErrorCode
    c_WriteParameter.errcheck = HandleErrorCode
    c_WriteParameter(writerHandle, self, name, parameterType)

def GlobalDefWriter_WriteRmaWin(writerHandle, self, name, comm, flags):
    c_WriteRmaWin = conf.lib.OTF2_GlobalDefWriter_WriteRmaWin
    c_WriteRmaWin.argtypes = [ ctypes.POINTER(GlobalDefWriter), RmaWinRef, StringRef, CommRef, RmaWinFlag ]
    c_WriteRmaWin.restype = ErrorCode
    c_WriteRmaWin.errcheck = HandleErrorCode
    c_WriteRmaWin(writerHandle, self, name, comm, flags)

def GlobalDefWriter_WriteMetricClassRecorder(writerHandle, metric, recorder):
    c_WriteMetricClassRecorder = conf.lib.OTF2_GlobalDefWriter_WriteMetricClassRecorder
    c_WriteMetricClassRecorder.argtypes = [ ctypes.POINTER(GlobalDefWriter), MetricRef, LocationRef ]
    c_WriteMetricClassRecorder.restype = ErrorCode
    c_WriteMetricClassRecorder.errcheck = HandleErrorCode
    c_WriteMetricClassRecorder(writerHandle, metric, recorder)

def GlobalDefWriter_WriteSystemTreeNodeProperty(writerHandle, systemTreeNode, name, type, value):
    c_WriteSystemTreeNodeProperty = conf.lib.OTF2_GlobalDefWriter_WriteSystemTreeNodeProperty
    c_WriteSystemTreeNodeProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), SystemTreeNodeRef, StringRef, Type, AttributeValue._arg_type ]
    c_WriteSystemTreeNodeProperty.restype = ErrorCode
    c_WriteSystemTreeNodeProperty.errcheck = HandleErrorCode
    c_WriteSystemTreeNodeProperty(writerHandle, systemTreeNode, name, type, value._arg_value)

def GlobalDefWriter_WriteSystemTreeNodeDomain(writerHandle, systemTreeNode, systemTreeDomain):
    c_WriteSystemTreeNodeDomain = conf.lib.OTF2_GlobalDefWriter_WriteSystemTreeNodeDomain
    c_WriteSystemTreeNodeDomain.argtypes = [ ctypes.POINTER(GlobalDefWriter), SystemTreeNodeRef, SystemTreeDomain ]
    c_WriteSystemTreeNodeDomain.restype = ErrorCode
    c_WriteSystemTreeNodeDomain.errcheck = HandleErrorCode
    c_WriteSystemTreeNodeDomain(writerHandle, systemTreeNode, systemTreeDomain)

def GlobalDefWriter_WriteLocationGroupProperty(writerHandle, locationGroup, name, type, value):
    c_WriteLocationGroupProperty = conf.lib.OTF2_GlobalDefWriter_WriteLocationGroupProperty
    c_WriteLocationGroupProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), LocationGroupRef, StringRef, Type, AttributeValue._arg_type ]
    c_WriteLocationGroupProperty.restype = ErrorCode
    c_WriteLocationGroupProperty.errcheck = HandleErrorCode
    c_WriteLocationGroupProperty(writerHandle, locationGroup, name, type, value._arg_value)

def GlobalDefWriter_WriteLocationProperty(writerHandle, location, name, type, value):
    c_WriteLocationProperty = conf.lib.OTF2_GlobalDefWriter_WriteLocationProperty
    c_WriteLocationProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), LocationRef, StringRef, Type, AttributeValue._arg_type ]
    c_WriteLocationProperty.restype = ErrorCode
    c_WriteLocationProperty.errcheck = HandleErrorCode
    c_WriteLocationProperty(writerHandle, location, name, type, value._arg_value)

def GlobalDefWriter_WriteCartDimension(writerHandle, self, name, size, cartPeriodicity):
    c_WriteCartDimension = conf.lib.OTF2_GlobalDefWriter_WriteCartDimension
    c_WriteCartDimension.argtypes = [ ctypes.POINTER(GlobalDefWriter), CartDimensionRef, StringRef, ctypes.c_uint32, CartPeriodicity ]
    c_WriteCartDimension.restype = ErrorCode
    c_WriteCartDimension.errcheck = HandleErrorCode
    c_WriteCartDimension(writerHandle, self, name, size, cartPeriodicity)

def GlobalDefWriter_WriteCartTopology(writerHandle, self, name, communicator, cartDimensions):
    numberOfDimensions = len(cartDimensions)
    array_type = CartDimensionRef * numberOfDimensions
    cart_dimensions_array = array_type()
    cart_dimensions_array[:] = cartDimensions

    c_WriteCartTopology = conf.lib.OTF2_GlobalDefWriter_WriteCartTopology
    c_WriteCartTopology.argtypes = [ ctypes.POINTER(GlobalDefWriter), CartTopologyRef, StringRef, CommRef,
                                     ctypes.c_uint8, array_type ]
    c_WriteCartTopology.restype = ErrorCode
    c_WriteCartTopology.errcheck = HandleErrorCode
    c_WriteCartTopology(writerHandle, self, name, communicator, numberOfDimensions, cart_dimensions_array)

def GlobalDefWriter_WriteCartCoordinate(writerHandle, cartTopology, rank, coordinates):
    numberOfDimensions = len(coordinates)
    array_type = ctypes.c_uint32 * numberOfDimensions
    coordinates_array = array_type()
    coordinates_array[:] = coordinates

    c_WriteCartCoordinate = conf.lib.OTF2_GlobalDefWriter_WriteCartCoordinate
    c_WriteCartCoordinate.argtypes = [ ctypes.POINTER(GlobalDefWriter), CartTopologyRef, ctypes.c_uint32,
                                       ctypes.c_uint8, array_type ]
    c_WriteCartCoordinate.restype = ErrorCode
    c_WriteCartCoordinate.errcheck = HandleErrorCode
    c_WriteCartCoordinate(writerHandle, cartTopology, rank, numberOfDimensions, coordinates_array)

def GlobalDefWriter_WriteSourceCodeLocation(writerHandle, self, file, lineNumber):
    c_WriteSourceCodeLocation = conf.lib.OTF2_GlobalDefWriter_WriteSourceCodeLocation
    c_WriteSourceCodeLocation.argtypes = [ ctypes.POINTER(GlobalDefWriter), SourceCodeLocationRef, StringRef, ctypes.c_uint32 ]
    c_WriteSourceCodeLocation.restype = ErrorCode
    c_WriteSourceCodeLocation.errcheck = HandleErrorCode
    c_WriteSourceCodeLocation(writerHandle, self, file, lineNumber)

def GlobalDefWriter_WriteCallingContext(writerHandle, self, region, sourceCodeLocation, parent):
    c_WriteCallingContext = conf.lib.OTF2_GlobalDefWriter_WriteCallingContext
    c_WriteCallingContext.argtypes = [ ctypes.POINTER(GlobalDefWriter), CallingContextRef, RegionRef, SourceCodeLocationRef, CallingContextRef ]
    c_WriteCallingContext.restype = ErrorCode
    c_WriteCallingContext.errcheck = HandleErrorCode
    c_WriteCallingContext(writerHandle, self, region, sourceCodeLocation, parent)

def GlobalDefWriter_WriteCallingContextProperty(writerHandle, callingContext, name, type, value):
    c_WriteCallingContextProperty = conf.lib.OTF2_GlobalDefWriter_WriteCallingContextProperty
    c_WriteCallingContextProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), CallingContextRef, StringRef, Type, AttributeValue._arg_type ]
    c_WriteCallingContextProperty.restype = ErrorCode
    c_WriteCallingContextProperty.errcheck = HandleErrorCode
    c_WriteCallingContextProperty(writerHandle, callingContext, name, type, value._arg_value)

def GlobalDefWriter_WriteInterruptGenerator(writerHandle, self, name, interruptGeneratorMode, base, exponent, period):
    c_WriteInterruptGenerator = conf.lib.OTF2_GlobalDefWriter_WriteInterruptGenerator
    c_WriteInterruptGenerator.argtypes = [ ctypes.POINTER(GlobalDefWriter), InterruptGeneratorRef, StringRef, InterruptGeneratorMode, Base, ctypes.c_int64, ctypes.c_uint64 ]
    c_WriteInterruptGenerator.restype = ErrorCode
    c_WriteInterruptGenerator.errcheck = HandleErrorCode
    c_WriteInterruptGenerator(writerHandle, self, name, interruptGeneratorMode, base, exponent, period)

def GlobalDefWriter_WriteIoFileProperty(writerHandle, ioFile, name, type, value):
    c_WriteIoFileProperty = conf.lib.OTF2_GlobalDefWriter_WriteIoFileProperty
    c_WriteIoFileProperty.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoFileRef, StringRef, Type, AttributeValue._arg_type ]
    c_WriteIoFileProperty.restype = ErrorCode
    c_WriteIoFileProperty.errcheck = HandleErrorCode
    c_WriteIoFileProperty(writerHandle, ioFile, name, type, value._arg_value)

def GlobalDefWriter_WriteIoRegularFile(writerHandle, self, name, scope):
    c_WriteIoRegularFile = conf.lib.OTF2_GlobalDefWriter_WriteIoRegularFile
    c_WriteIoRegularFile.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoFileRef, StringRef, SystemTreeNodeRef ]
    c_WriteIoRegularFile.restype = ErrorCode
    c_WriteIoRegularFile.errcheck = HandleErrorCode
    c_WriteIoRegularFile(writerHandle, self, name, scope)

def GlobalDefWriter_WriteIoDirectory(writerHandle, self, name, scope):
    c_WriteIoDirectory = conf.lib.OTF2_GlobalDefWriter_WriteIoDirectory
    c_WriteIoDirectory.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoFileRef, StringRef, SystemTreeNodeRef ]
    c_WriteIoDirectory.restype = ErrorCode
    c_WriteIoDirectory.errcheck = HandleErrorCode
    c_WriteIoDirectory(writerHandle, self, name, scope)

def GlobalDefWriter_WriteIoHandle(writerHandle, self, name, file, ioParadigm, ioHandleFlags, comm, parent):
    c_WriteIoHandle = conf.lib.OTF2_GlobalDefWriter_WriteIoHandle
    c_WriteIoHandle.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoHandleRef, StringRef, IoFileRef, IoParadigmRef, IoHandleFlag, CommRef, IoHandleRef ]
    c_WriteIoHandle.restype = ErrorCode
    c_WriteIoHandle.errcheck = HandleErrorCode
    c_WriteIoHandle(writerHandle, self, name, file, ioParadigm, ioHandleFlags, comm, parent)

def GlobalDefWriter_WriteIoPreCreatedHandleState(writerHandle, ioHandle, mode, statusFlags):
    c_WriteIoPreCreatedHandleState = conf.lib.OTF2_GlobalDefWriter_WriteIoPreCreatedHandleState
    c_WriteIoPreCreatedHandleState.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoHandleRef, IoAccessMode, IoStatusFlag ]
    c_WriteIoPreCreatedHandleState.restype = ErrorCode
    c_WriteIoPreCreatedHandleState.errcheck = HandleErrorCode
    c_WriteIoPreCreatedHandleState(writerHandle, ioHandle, mode, statusFlags)

def GlobalDefWriter_WriteCallpathParameter(writerHandle, callpath, parameter, type, value):
    c_WriteCallpathParameter = conf.lib.OTF2_GlobalDefWriter_WriteCallpathParameter
    c_WriteCallpathParameter.argtypes = [ ctypes.POINTER(GlobalDefWriter), CallpathRef, ParameterRef, Type, AttributeValue._arg_type ]
    c_WriteCallpathParameter.restype = ErrorCode
    c_WriteCallpathParameter.errcheck = HandleErrorCode
    c_WriteCallpathParameter(writerHandle, callpath, parameter, type, value._arg_value)

def GlobalDefWriter_WriteInterComm(writerHandle, self, name, groupA, groupB, commonCommunicator, flags):
    c_WriteInterComm = conf.lib.OTF2_GlobalDefWriter_WriteInterComm
    c_WriteInterComm.argtypes = [ ctypes.POINTER(GlobalDefWriter), CommRef, StringRef, GroupRef, GroupRef, CommRef, CommFlag ]
    c_WriteInterComm.restype = ErrorCode
    c_WriteInterComm.errcheck = HandleErrorCode
    c_WriteInterComm(writerHandle, self, name, groupA, groupB, commonCommunicator, flags)

__all__ = [
    'GlobalDefWriter',
    'GlobalDefWriter_GetNumberOfDefinitions',
    'GlobalDefWriter_GetNumberOfLocations',
    'GlobalDefWriter_WriteClockProperties',
    'GlobalDefWriter_WriteParadigm',
    'GlobalDefWriter_WriteParadigmProperty',
    'GlobalDefWriter_WriteIoParadigm',
    'GlobalDefWriter_WriteString',
    'GlobalDefWriter_WriteAttribute',
    'GlobalDefWriter_WriteSystemTreeNode',
    'GlobalDefWriter_WriteLocationGroup',
    'GlobalDefWriter_WriteLocation',
    'GlobalDefWriter_WriteRegion',
    'GlobalDefWriter_WriteCallsite',
    'GlobalDefWriter_WriteCallpath',
    'GlobalDefWriter_WriteGroup',
    'GlobalDefWriter_WriteMetricMember',
    'GlobalDefWriter_WriteMetricClass',
    'GlobalDefWriter_WriteMetricInstance',
    'GlobalDefWriter_WriteComm',
    'GlobalDefWriter_WriteParameter',
    'GlobalDefWriter_WriteRmaWin',
    'GlobalDefWriter_WriteMetricClassRecorder',
    'GlobalDefWriter_WriteSystemTreeNodeProperty',
    'GlobalDefWriter_WriteSystemTreeNodeDomain',
    'GlobalDefWriter_WriteLocationGroupProperty',
    'GlobalDefWriter_WriteLocationProperty',
    'GlobalDefWriter_WriteCartDimension',
    'GlobalDefWriter_WriteCartTopology',
    'GlobalDefWriter_WriteCartCoordinate',
    'GlobalDefWriter_WriteSourceCodeLocation',
    'GlobalDefWriter_WriteCallingContext',
    'GlobalDefWriter_WriteCallingContextProperty',
    'GlobalDefWriter_WriteInterruptGenerator',
    'GlobalDefWriter_WriteIoFileProperty',
    'GlobalDefWriter_WriteIoRegularFile',
    'GlobalDefWriter_WriteIoDirectory',
    'GlobalDefWriter_WriteIoHandle',
    'GlobalDefWriter_WriteIoPreCreatedHandleState',
    'GlobalDefWriter_WriteCallpathParameter',
    'GlobalDefWriter_WriteInterComm',
]



import os
import os.path
import unittest
from shutil import rmtree
from tempfile import mkdtemp

import otf2


TIMER_GRANULARITY = 1000000


def t():
    return int(round(time.time() * TIMER_GRANULARITY))


class TestOTF2Groups(unittest.TestCase):
    def setUp(self):
        self.old_cwd = os.getcwd()
        self.tmp_dirname = mkdtemp()
        self.tmp_dirname = mkdtemp(prefix=os.path.basename(os.path.abspath(__file__))[:-3] + '_tmp', dir=self.old_cwd)

        self.orig_trace = os.path.join(self.tmp_dirname, "orig")
        self.retrace = os.path.join(self.tmp_dirname, "retrc")

        os.mkdir(self.orig_trace)
        os.mkdir(self.retrace)

        os.chdir(self.tmp_dirname)

    def tearDown(self):
        os.chdir(self.old_cwd)
        if os.getenv('KEEP_TEST_OUTPUT', '') != '':
            print(self.tmp_dirname)
        else:
            rmtree(self.tmp_dirname)

    num_locations = 8
    paradigm = otf2.Paradigm.MPI
    group_ranks = range(num_locations-1, num_locations//2-1, -1)

    def generate_trace(self, archive_name):
        with otf2.writer.open(archive_name, timer_resolution=TIMER_GRANULARITY) as trace:
            self.definitions = trace.definitions

            system_tree_node = trace.definitions.system_tree_node("Root", parent=None)
            location_group = trace.definitions.location_group("Process",
                                                              system_tree_parent=system_tree_node)
            locations = [trace.definitions.location("Location {}".format(rank),
                                                    group=location_group)
                         for rank in range(self.num_locations)]
            writers = [trace.event_writer_from_location(location) for location in locations]

            mpi_world = trace.definitions.group("MPI_COMM_WORLD",
                                                paradigm=self.paradigm,
                                                group_type=otf2.GroupType.COMM_LOCATIONS,
                                                members=locations)

            group1 = trace.definitions.group("MPI_Group1",
                                             paradigm=self.paradigm,
                                             group_type=otf2.GroupType.COMM_GROUP,
                                             members=self.group_ranks)

            group2 = trace.definitions.group("MPI_Group2",
                                             paradigm=self.paradigm,
                                             group_type=otf2.GroupType.COMM_GROUP,
                                             members=[locations[rank] for rank in
                                                      self.group_ranks])

            self.assert_groups_equal(mpi_world, group1, group2)

    def read_trace(self, archive_name):
        with otf2.reader.open(archive_name) as trace:
            self.assertEqual(3, len(trace.definitions.groups))
            mpi_world = [group for group in trace.definitions.groups if group.group_type ==
                         otf2.GroupType.COMM_LOCATIONS]
            self.assertEqual(1, len(mpi_world))
            mpi_world, = mpi_world
            comm_groups = [group for group in trace.definitions.groups if group.group_type ==
                           otf2.GroupType.COMM_GROUP]
            self.assertEqual(2, len(comm_groups))
            self.assert_groups_equal(mpi_world, *comm_groups)

    def rewrite_trace(self, old_anchor_path, new_archive_path):
        with otf2.reader.open(old_anchor_path) as trace_reader:
            with otf2.writer.open(new_archive_path,
                                  definitions=trace_reader.definitions) as write_trace:
                for location, event in trace_reader.events:
                    writer = write_trace.event_writer_from_location(location)
                    writer(event)

    def assert_groups_equal(self, world, group1, group2):
            for global_rank, local_rank, l1, l2 in \
                    zip(self.group_ranks, range(len(self.group_ranks)), group1.members, group2.members):
                self.assertIs(l1, l2)
                self.assertIs(l1, group1.location(local_rank))
                self.assertIs(l1, group2.location(local_rank))
                self.assertEqual(local_rank, group1.rank(l1))
                self.assertEqual(local_rank, group2.rank(l1))

                self.assertIs(l1, world.location(global_rank))
                self.assertIs(global_rank, world.rank(l1))

    def test_rewrite(self):
        self.generate_trace(self.orig_trace)
        self.rewrite_trace(os.path.join(self.orig_trace, "traces.otf2"), self.retrace)
        self.read_trace(os.path.join(self.retrace, "traces.otf2"))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# This file is part of the Score-P software (http://www.score-p.org)
#
# Copyright (c) 2015-2016, 2021,
# Technische Universitaet Dresden, Germany
#
# This software may be modified and distributed under the terms of
# a BSD-style license.  See the COPYING file in the package base
# directory for details.
#

import otf2
from otf2.enums import Type
import time


TIMER_GRANULARITY = 1000000


def t():
    return int(round(time.time() * TIMER_GRANULARITY))


with otf2.writer.open("TestArchive", timer_resolution=TIMER_GRANULARITY) as trace:

    function = trace.definitions.region("My Function")

    parent_node = trace.definitions.system_tree_node("node")
    system_tree_node = trace.definitions.system_tree_node("myHost", parent=parent_node)

    trace.definitions.system_tree_node_property(system_tree_node, "color", value="black")
    trace.definitions.system_tree_node_property(system_tree_node, "rack #", value=42)

    location_group = trace.definitions.location_group("Initial Process",
                                                      system_tree_parent=system_tree_node)

    attr = trace.definitions.attribute("StringTest", "A test attribute", Type.STRING)
    float_attr = trace.definitions.attribute("FloatTest", "Another test attribute",
                                             Type.DOUBLE)

    writer = trace.event_writer("Main Thread", group=location_group)

    # Write enter and leave event
    writer.enter(t(), function, {attr: "Hello World"})
    writer.leave(t(), function, attributes={float_attr: 42.0, attr: "Wurst?"})

    # Get convenience metric object and write one metric event
    temperature = trace.definitions.metric("Time since last coffee", unit="min")
    writer.metric(t(), temperature, 72.0)

    # Get metric members
    temp_member = trace.definitions.metric_member("Temperature", "C", otf2.MetricType.OTHER,
                                                  otf2.MetricMode.ABSOLUTE_POINT)
    power_member = trace.definitions.metric_member("Power", "W")
    # Add metric members to the metric class object
    mclass = trace.definitions.metric_class([temp_member, power_member])
    # Add metric object to the location object
    writer.metric(t(), mclass, [42.0, 12345.6])



#!/usr/bin/env python

import otf2
import sys


def rewrite(old_anchor_path, new_archive_path):
    with otf2.reader.open(old_anchor_path) as trace_reader:
        with otf2.writer.open(new_archive_path,
                              definitions=trace_reader.definitions) as write_trace:
            for location, event in trace_reader.events:
                writer = write_trace.event_writer_from_location(location)
                event.time -= trace_reader.definitions.clock_properties.global_offset
                writer(event)


if __name__ == '__main__':
    rewrite(sys.argv[1], sys.argv[2])




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import EnumBase, EnumBitset
from .GeneralDefinitions import UNDEFINED_UINT32

class MeasurementMode(EnumBase, ctypes.c_uint8):
    pass

MEASUREMENT_ON = MeasurementMode._construct(1, "ON")
MEASUREMENT_OFF = MeasurementMode._construct(2, "OFF")

class CollectiveOp(EnumBase, ctypes.c_uint8):
    pass

COLLECTIVE_OP_BARRIER = CollectiveOp._construct(0, "BARRIER")
COLLECTIVE_OP_BCAST = CollectiveOp._construct(1, "BCAST")
COLLECTIVE_OP_GATHER = CollectiveOp._construct(2, "GATHER")
COLLECTIVE_OP_GATHERV = CollectiveOp._construct(3, "GATHERV")
COLLECTIVE_OP_SCATTER = CollectiveOp._construct(4, "SCATTER")
COLLECTIVE_OP_SCATTERV = CollectiveOp._construct(5, "SCATTERV")
COLLECTIVE_OP_ALLGATHER = CollectiveOp._construct(6, "ALLGATHER")
COLLECTIVE_OP_ALLGATHERV = CollectiveOp._construct(7, "ALLGATHERV")
COLLECTIVE_OP_ALLTOALL = CollectiveOp._construct(8, "ALLTOALL")
COLLECTIVE_OP_ALLTOALLV = CollectiveOp._construct(9, "ALLTOALLV")
COLLECTIVE_OP_ALLTOALLW = CollectiveOp._construct(10, "ALLTOALLW")
COLLECTIVE_OP_ALLREDUCE = CollectiveOp._construct(11, "ALLREDUCE")
COLLECTIVE_OP_REDUCE = CollectiveOp._construct(12, "REDUCE")
COLLECTIVE_OP_REDUCE_SCATTER = CollectiveOp._construct(13, "REDUCE_SCATTER")
COLLECTIVE_OP_SCAN = CollectiveOp._construct(14, "SCAN")
COLLECTIVE_OP_EXSCAN = CollectiveOp._construct(15, "EXSCAN")
COLLECTIVE_OP_REDUCE_SCATTER_BLOCK = CollectiveOp._construct(16, "REDUCE_SCATTER_BLOCK")
COLLECTIVE_OP_CREATE_HANDLE = CollectiveOp._construct(17, "CREATE_HANDLE")
COLLECTIVE_OP_DESTROY_HANDLE = CollectiveOp._construct(18, "DESTROY_HANDLE")
COLLECTIVE_OP_ALLOCATE = CollectiveOp._construct(19, "ALLOCATE")
COLLECTIVE_OP_DEALLOCATE = CollectiveOp._construct(20, "DEALLOCATE")
COLLECTIVE_OP_CREATE_HANDLE_AND_ALLOCATE = CollectiveOp._construct(21, "CREATE_HANDLE_AND_ALLOCATE")
COLLECTIVE_OP_DESTROY_HANDLE_AND_DEALLOCATE = CollectiveOp._construct(22, "DESTROY_HANDLE_AND_DEALLOCATE")

class RmaSyncType(EnumBase, ctypes.c_uint8):
    pass

RMA_SYNC_TYPE_MEMORY = RmaSyncType._construct(0, "MEMORY")
RMA_SYNC_TYPE_NOTIFY_IN = RmaSyncType._construct(1, "NOTIFY_IN")
RMA_SYNC_TYPE_NOTIFY_OUT = RmaSyncType._construct(2, "NOTIFY_OUT")

class RmaSyncLevel(EnumBitset, ctypes.c_uint32):
    pass

RMA_SYNC_LEVEL_NONE = RmaSyncLevel._construct(0, "NONE")
RMA_SYNC_LEVEL_PROCESS = RmaSyncLevel._construct(( 1 << 0 ), "PROCESS")
RMA_SYNC_LEVEL_MEMORY = RmaSyncLevel._construct(( 1 << 1 ), "MEMORY")

class LockType(EnumBase, ctypes.c_uint8):
    pass

LOCK_EXCLUSIVE = LockType._construct(0, "EXCLUSIVE")
LOCK_SHARED = LockType._construct(1, "SHARED")

class RmaAtomicType(EnumBase, ctypes.c_uint8):
    pass

RMA_ATOMIC_TYPE_ACCUMULATE = RmaAtomicType._construct(0, "ACCUMULATE")
RMA_ATOMIC_TYPE_INCREMENT = RmaAtomicType._construct(1, "INCREMENT")
RMA_ATOMIC_TYPE_TEST_AND_SET = RmaAtomicType._construct(2, "TEST_AND_SET")
RMA_ATOMIC_TYPE_COMPARE_AND_SWAP = RmaAtomicType._construct(3, "COMPARE_AND_SWAP")
RMA_ATOMIC_TYPE_SWAP = RmaAtomicType._construct(4, "SWAP")
RMA_ATOMIC_TYPE_FETCH_AND_ADD = RmaAtomicType._construct(5, "FETCH_AND_ADD")
RMA_ATOMIC_TYPE_FETCH_AND_INCREMENT = RmaAtomicType._construct(6, "FETCH_AND_INCREMENT")
RMA_ATOMIC_TYPE_FETCH_AND_ACCUMULATE = RmaAtomicType._construct(7, "FETCH_AND_ACCUMULATE")

class IoCreationFlag(EnumBitset, ctypes.c_uint32):
    pass

IO_CREATION_FLAG_NONE = IoCreationFlag._construct(0, "NONE")
IO_CREATION_FLAG_CREATE = IoCreationFlag._construct(1 << 0, "CREATE")
IO_CREATION_FLAG_TRUNCATE = IoCreationFlag._construct(1 << 1, "TRUNCATE")
IO_CREATION_FLAG_DIRECTORY = IoCreationFlag._construct(1 << 2, "DIRECTORY")
IO_CREATION_FLAG_EXCLUSIVE = IoCreationFlag._construct(1 << 3, "EXCLUSIVE")
IO_CREATION_FLAG_NO_CONTROLLING_TERMINAL = IoCreationFlag._construct(1 << 4, "NO_CONTROLLING_TERMINAL")
IO_CREATION_FLAG_NO_FOLLOW = IoCreationFlag._construct(1 << 5, "NO_FOLLOW")
IO_CREATION_FLAG_PATH = IoCreationFlag._construct(1 << 6, "PATH")
IO_CREATION_FLAG_TEMPORARY_FILE = IoCreationFlag._construct(1 <<7, "TEMPORARY_FILE")
IO_CREATION_FLAG_LARGEFILE = IoCreationFlag._construct(1 << 8, "LARGEFILE")
IO_CREATION_FLAG_NO_SEEK = IoCreationFlag._construct(1 << 9, "NO_SEEK")
IO_CREATION_FLAG_UNIQUE = IoCreationFlag._construct(1 << 10, "UNIQUE")

class IoSeekOption(EnumBase, ctypes.c_uint8):
    pass

IO_SEEK_FROM_START = IoSeekOption._construct(0, "FROM_START")
IO_SEEK_FROM_CURRENT = IoSeekOption._construct(1, "FROM_CURRENT")
IO_SEEK_FROM_END = IoSeekOption._construct(2, "FROM_END")
IO_SEEK_DATA = IoSeekOption._construct(3, "DATA")
IO_SEEK_HOLE = IoSeekOption._construct(4, "HOLE")

class IoOperationMode(EnumBase, ctypes.c_uint8):
    pass

IO_OPERATION_MODE_READ = IoOperationMode._construct(0, "READ")
IO_OPERATION_MODE_WRITE = IoOperationMode._construct(1, "WRITE")
IO_OPERATION_MODE_FLUSH = IoOperationMode._construct(2, "FLUSH")

class IoOperationFlag(EnumBitset, ctypes.c_uint32):
    pass

IO_OPERATION_FLAG_NONE = IoOperationFlag._construct(0, "NONE")
IO_OPERATION_FLAG_NON_BLOCKING = IoOperationFlag._construct(1 << 0, "NON_BLOCKING")
IO_OPERATION_FLAG_COLLECTIVE = IoOperationFlag._construct(1 << 1, "COLLECTIVE")

class CollectiveRoot(EnumBase, ctypes.c_uint32):
    pass


class MetricValue(ctypes.Union):
    _fields_ = [
        ("signed_int", ctypes.c_int64),
        ("unsigned_int", ctypes.c_uint64),
        ("floating_point", ctypes.c_double),
    ]

COLLECTIVE_ROOT_NONE = CollectiveRoot._construct(UNDEFINED_UINT32.value, "NONE")
COLLECTIVE_ROOT_SELF = CollectiveRoot._construct(UNDEFINED_UINT32.value - 1, "SELF")
COLLECTIVE_ROOT_THIS_GROUP = CollectiveRoot._construct(UNDEFINED_UINT32.value - 2, "THIS_GROUP")

__all__ = [
    'MeasurementMode',
    'MEASUREMENT_ON',
    'MEASUREMENT_OFF',
    'CollectiveOp',
    'COLLECTIVE_OP_BARRIER',
    'COLLECTIVE_OP_BCAST',
    'COLLECTIVE_OP_GATHER',
    'COLLECTIVE_OP_GATHERV',
    'COLLECTIVE_OP_SCATTER',
    'COLLECTIVE_OP_SCATTERV',
    'COLLECTIVE_OP_ALLGATHER',
    'COLLECTIVE_OP_ALLGATHERV',
    'COLLECTIVE_OP_ALLTOALL',
    'COLLECTIVE_OP_ALLTOALLV',
    'COLLECTIVE_OP_ALLTOALLW',
    'COLLECTIVE_OP_ALLREDUCE',
    'COLLECTIVE_OP_REDUCE',
    'COLLECTIVE_OP_REDUCE_SCATTER',
    'COLLECTIVE_OP_SCAN',
    'COLLECTIVE_OP_EXSCAN',
    'COLLECTIVE_OP_REDUCE_SCATTER_BLOCK',
    'COLLECTIVE_OP_CREATE_HANDLE',
    'COLLECTIVE_OP_DESTROY_HANDLE',
    'COLLECTIVE_OP_ALLOCATE',
    'COLLECTIVE_OP_DEALLOCATE',
    'COLLECTIVE_OP_CREATE_HANDLE_AND_ALLOCATE',
    'COLLECTIVE_OP_DESTROY_HANDLE_AND_DEALLOCATE',
    'RmaSyncType',
    'RMA_SYNC_TYPE_MEMORY',
    'RMA_SYNC_TYPE_NOTIFY_IN',
    'RMA_SYNC_TYPE_NOTIFY_OUT',
    'RmaSyncLevel',
    'RMA_SYNC_LEVEL_NONE',
    'RMA_SYNC_LEVEL_PROCESS',
    'RMA_SYNC_LEVEL_MEMORY',
    'LockType',
    'LOCK_EXCLUSIVE',
    'LOCK_SHARED',
    'RmaAtomicType',
    'RMA_ATOMIC_TYPE_ACCUMULATE',
    'RMA_ATOMIC_TYPE_INCREMENT',
    'RMA_ATOMIC_TYPE_TEST_AND_SET',
    'RMA_ATOMIC_TYPE_COMPARE_AND_SWAP',
    'RMA_ATOMIC_TYPE_SWAP',
    'RMA_ATOMIC_TYPE_FETCH_AND_ADD',
    'RMA_ATOMIC_TYPE_FETCH_AND_INCREMENT',
    'RMA_ATOMIC_TYPE_FETCH_AND_ACCUMULATE',
    'IoCreationFlag',
    'IO_CREATION_FLAG_NONE',
    'IO_CREATION_FLAG_CREATE',
    'IO_CREATION_FLAG_TRUNCATE',
    'IO_CREATION_FLAG_DIRECTORY',
    'IO_CREATION_FLAG_EXCLUSIVE',
    'IO_CREATION_FLAG_NO_CONTROLLING_TERMINAL',
    'IO_CREATION_FLAG_NO_FOLLOW',
    'IO_CREATION_FLAG_PATH',
    'IO_CREATION_FLAG_TEMPORARY_FILE',
    'IO_CREATION_FLAG_LARGEFILE',
    'IO_CREATION_FLAG_NO_SEEK',
    'IO_CREATION_FLAG_UNIQUE',
    'IoSeekOption',
    'IO_SEEK_FROM_START',
    'IO_SEEK_FROM_CURRENT',
    'IO_SEEK_FROM_END',
    'IO_SEEK_DATA',
    'IO_SEEK_HOLE',
    'IoOperationMode',
    'IO_OPERATION_MODE_READ',
    'IO_OPERATION_MODE_WRITE',
    'IO_OPERATION_MODE_FLUSH',
    'IoOperationFlag',
    'IO_OPERATION_FLAG_NONE',
    'IO_OPERATION_FLAG_NON_BLOCKING',
    'IO_OPERATION_FLAG_COLLECTIVE',
    'CollectiveRoot',
    'COLLECTIVE_ROOT_NONE',
    'COLLECTIVE_ROOT_SELF',
    'COLLECTIVE_ROOT_THIS_GROUP',
    'MetricValue'
]



import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GlobalSnapReaderCallbacks import GlobalSnapReaderCallbacks


class GlobalSnapReader(ctypes.Structure):
    pass

def GlobalSnapReader_SetCallbacks(reader, callbacks, userData):
    c_SetCallbacks = conf.lib.OTF2_GlobalSnapReader_SetCallbacks
    c_SetCallbacks.argtypes = [ ctypes.POINTER(GlobalSnapReader), ctypes.POINTER(GlobalSnapReaderCallbacks), ctypes.py_object ]
    c_SetCallbacks.restype = ErrorCode
    c_SetCallbacks.errcheck = HandleErrorCode
    c_SetCallbacks(reader, callbacks, ctypes.py_object(userData))


def GlobalSnapReader_ReadSnapshots(reader, recordsToRead):
    c_ReadSnapshots = conf.lib.OTF2_GlobalSnapReader_ReadSnapshots
    c_ReadSnapshots.argtypes = [ ctypes.POINTER(GlobalSnapReader), ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint64) ]
    c_ReadSnapshots.restype = ErrorCode
    c_ReadSnapshots.errcheck = HandleErrorCode
    recordsRead = ctypes.c_uint64()
    c_ReadSnapshots(reader, recordsToRead, ctypes.byref(recordsRead))
    return recordsRead.value

__all__ = [
    'GlobalSnapReader',
    'GlobalSnapReader_ReadSnapshots',
    'GlobalSnapReader_SetCallbacks'
]



from __future__ import print_function

import sys
import os

try:
    from hashlib import md5
except:
    from md5 import new as md5

if len(sys.argv) == 1:
    print('Invalid number of arguments', file=sys.stderr)
    sys.exit(3)

m = md5()
hexdigest_length = m.digest_size * 2

verbose = False
if sys.argv[1] in ('--verbose'):
    verbose = True
    sys.argv = sys.argv[:1] + sys.argv[2:]

if sys.argv[1] in ('--check', '--check-output', '--check-input'):
    if len(sys.argv) != 3:
        print('Invalid number of arguments', file=sys.stderr)
        sys.exit(3)

    # next arg is sig file
    signame = sys.argv[2]

    # read the sigfile and the two signatures for the output and input
    fp = open(signame)
    sig = fp.read()
    fp.close()
    outsig = sig[0:hexdigest_length]
    insig  = sig[hexdigest_length+1:2 * hexdigest_length+1]
    filenames = [x for x in sig[2 * hexdigest_length + 2:].split('\n') if x]

    sigdir = os.path.dirname(signame)
    # than the output file
    outname = filenames[0]
    outfilepath = os.path.normpath(os.path.join(sigdir, outname))
    # remaings are the input files
    infiles = filenames[1:]

    if sys.argv[1] in ('--check', '--check-output'):
        fp = open(outfilepath)
        m.update(fp.read().encode())
        fp.close()
        if m.hexdigest() != outsig:
            if verbose:
                print('%s: template result does not match signature, please regenerate' % (outfilepath), file=sys.stderr)
            sys.exit(1)
        m = md5()

    if sys.argv[1] in ('--check', '--check-input'):
        for name in infiles:
            filepath = os.path.normpath(os.path.join(sigdir, name))
            fp = open(filepath)
            m.update(fp.read().encode())
            fp.close()
        if m.hexdigest() != insig:
            if verbose:
                print('%s: template inputs do not match with signature, please regenerate' % (outfilepath,), file=sys.stderr)
            sys.exit(2)
        m = md5()

    sys.exit(0)

# generate mode

if len(sys.argv) < 2:
    print('Invalid number of arguments', file=sys.stderr)
    sys.exit(3)

# second arg is output name
#print('out: %s' % (sys.argv[2],), file=sys.stderr)
fp = open(sys.argv[1])
m.update(fp.read().encode())
fp.close()
outsig = m.hexdigest()
m = md5()

# remaining files are input
for name in sys.argv[2:]:
    #print('in:  %s' % (name,), file=sys.stderr)
    fp = open(name)
    m.update(fp.read().encode())
    fp.close()
insig = m.hexdigest()

# write sig file
sys.stdout.write('%s\n%s\n' % (outsig, insig))
for name in sys.argv[1:]:
    sys.stdout.write('%s\n' % (name,))

sys.exit(0)



# Generated by OTF2 Template Engine

import _otf2.GlobalDefWriter
from .definitions import *


class DefinitionWriter(object):
    """
        This class wraps the OTF2_GlobalDefWriter.

        Used internally only.
    """
    def __init__(self, archive):
        assert archive
        self._handle = _otf2.Archive_GetGlobalDefWriter(archive._handle)
        self._archive = archive

    def __enter__(self):
        assert self._handle
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        """Destructor.  Calls close()."""
        # The try/except block is in case this is called at program
        # exit time, when it's possible that globals have already been
        # deleted, and then the close() call might fail.  Since
        # there's nothing we can do about such failures and they annoy
        # the end users, we suppress the traceback.
        try:
            self.close()
        except:
            pass

    def close(self):
        if self._handle:
            try:
                _otf2.Archive_CloseGlobalDefWriter(self._archive._handle, self._handle)
            finally:
                self._archive = None
                self._handle = None

    @property
    def handle(self):
        """
            Gives access to the underlaying OTF2 global definition writer
        """
        if self._handle is None:
            raise TraceWriterError(
                "Trying to access handle, but GlobalDefWriter is already closed.")
        return self._handle

    @property
    def number_of_definitions(self):
        return _otf2.GlobalDefWriter.GetNumberOfDefinitions(self._handle)

    @property
    def number_of_locations(self):
        return _otf2.GlobalDefWriter.GetNumberOfLocations(self._handle)

    def write(self, definition, *args):
        {
            Paradigm: self.write_paradigm,
            ParadigmProperty: self.write_paradigm_property,
            IoParadigm: self.write_io_paradigm,
            String: self.write_string,
            Attribute: self.write_attribute,
            SystemTreeNode: self.write_system_tree_node,
            LocationGroup: self.write_location_group,
            Location: self.write_location,
            Region: self.write_region,
            Callsite: self.write_callsite,
            Callpath: self.write_callpath,
            Group: self.write_group,
            MetricMember: self.write_metric_member,
            MetricClass: self.write_metric_class,
            MetricInstance: self.write_metric_instance,
            Comm: self.write_comm,
            Parameter: self.write_parameter,
            RmaWin: self.write_rma_win,
            MetricClassRecorder: self.write_metric_class_recorder,
            SystemTreeNodeProperty: self.write_system_tree_node_property,
            SystemTreeNodeDomain: self.write_system_tree_node_domain,
            LocationGroupProperty: self.write_location_group_property,
            LocationProperty: self.write_location_property,
            CartDimension: self.write_cart_dimension,
            CartTopology: self.write_cart_topology,
            CartCoordinate: self.write_cart_coordinate,
            SourceCodeLocation: self.write_source_code_location,
            CallingContext: self.write_calling_context,
            CallingContextProperty: self.write_calling_context_property,
            InterruptGenerator: self.write_interrupt_generator,
            IoFileProperty: self.write_io_file_property,
            IoRegularFile: self.write_io_regular_file,
            IoDirectory: self.write_io_directory,
            IoHandle: self.write_io_handle,
            IoPreCreatedHandleState: self.write_io_pre_created_handle_state,
            CallpathParameter: self.write_callpath_parameter,
            InterComm: self.write_inter_comm,
        }[type(definition)](*args)

    def write_clock_properties(this, timerResolution, globalOffset, traceLength, realtimeTimestamp):
        _otf2.GlobalDefWriter_WriteClockProperties(this._handle, timerResolution, globalOffset, traceLength, realtimeTimestamp)

    def write_paradigm(this, paradigm, name, paradigmClass):
        _otf2.GlobalDefWriter_WriteParadigm(this._handle, paradigm, name, paradigmClass)

    def write_paradigm_property(this, paradigm, property, type, value):
        _otf2.GlobalDefWriter_WriteParadigmProperty(this._handle, paradigm, property, type, value)

    def write_io_paradigm(this, self, identification, name, ioParadigmClass, ioParadigmFlags, properties, types, values):
        _otf2.GlobalDefWriter_WriteIoParadigm(this._handle, self, identification, name,
                                              ioParadigmClass, ioParadigmFlags, properties, types, values)

    def write_string(this, self, string):
        _otf2.GlobalDefWriter_WriteString(this._handle, self, string)

    def write_attribute(this, self, name, description, type):
        _otf2.GlobalDefWriter_WriteAttribute(this._handle, self, name, description, type)

    def write_system_tree_node(this, self, name, className, parent):
        _otf2.GlobalDefWriter_WriteSystemTreeNode(this._handle, self, name, className, parent)

    def write_location_group(this, self, name, locationGroupType, systemTreeParent, creatingLocationGroup):
        _otf2.GlobalDefWriter_WriteLocationGroup(this._handle, self, name, locationGroupType, systemTreeParent, creatingLocationGroup)

    def write_location(this, self, name, locationType, numberOfEvents, locationGroup):
        _otf2.GlobalDefWriter_WriteLocation(this._handle, self, name, locationType, numberOfEvents, locationGroup)

    def write_region(this, self, name, canonicalName, description, regionRole, paradigm, regionFlags, sourceFile, beginLineNumber, endLineNumber):
        _otf2.GlobalDefWriter_WriteRegion(this._handle, self, name, canonicalName, description, regionRole, paradigm, regionFlags, sourceFile, beginLineNumber, endLineNumber)

    def write_callsite(this, self, sourceFile, lineNumber, enteredRegion, leftRegion):
        _otf2.GlobalDefWriter_WriteCallsite(this._handle, self, sourceFile, lineNumber, enteredRegion, leftRegion)

    def write_callpath(this, self, parent, region):
        _otf2.GlobalDefWriter_WriteCallpath(this._handle, self, parent, region)

    def write_group(this, self, name, groupType, paradigm, groupFlags, members):
        _otf2.GlobalDefWriter_WriteGroup(this._handle, self, name, groupType, paradigm, groupFlags, members)

    def write_metric_member(this, self, name, description, metricType, metricMode, valueType, base, exponent, unit):
        _otf2.GlobalDefWriter_WriteMetricMember(this._handle, self, name, description, metricType, metricMode, valueType, base, exponent, unit)

    def write_metric_class(this, self, metricMembers, metricOccurrence, recorderKind):
        _otf2.GlobalDefWriter_WriteMetricClass(this._handle, self, metricMembers, metricOccurrence, recorderKind)

    def write_metric_instance(this, self, metricClass, recorder, metricScope, scope):
        _otf2.GlobalDefWriter_WriteMetricInstance(this._handle, self, metricClass, recorder, metricScope, scope)

    def write_comm(this, self, name, group, parent, flags):
        _otf2.GlobalDefWriter_WriteComm(this._handle, self, name, group, parent, flags)

    def write_parameter(this, self, name, parameterType):
        _otf2.GlobalDefWriter_WriteParameter(this._handle, self, name, parameterType)

    def write_rma_win(this, self, name, comm, flags):
        _otf2.GlobalDefWriter_WriteRmaWin(this._handle, self, name, comm, flags)

    def write_metric_class_recorder(this, metric, recorder):
        _otf2.GlobalDefWriter_WriteMetricClassRecorder(this._handle, metric, recorder)

    def write_system_tree_node_property(this, systemTreeNode, name, type, value):
        _otf2.GlobalDefWriter_WriteSystemTreeNodeProperty(this._handle, systemTreeNode, name, type, value)

    def write_system_tree_node_domain(this, systemTreeNode, systemTreeDomain):
        _otf2.GlobalDefWriter_WriteSystemTreeNodeDomain(this._handle, systemTreeNode, systemTreeDomain)

    def write_location_group_property(this, locationGroup, name, type, value):
        _otf2.GlobalDefWriter_WriteLocationGroupProperty(this._handle, locationGroup, name, type, value)

    def write_location_property(this, location, name, type, value):
        _otf2.GlobalDefWriter_WriteLocationProperty(this._handle, location, name, type, value)

    def write_cart_dimension(this, self, name, size, cartPeriodicity):
        _otf2.GlobalDefWriter_WriteCartDimension(this._handle, self, name, size, cartPeriodicity)

    def write_cart_topology(this, self, name, communicator, cartDimensions):
        _otf2.GlobalDefWriter_WriteCartTopology(this._handle, self, name, communicator, cartDimensions)

    def write_cart_coordinate(this, cartTopology, rank, coordinates):
        _otf2.GlobalDefWriter_WriteCartCoordinate(this._handle, cartTopology, rank, coordinates)

    def write_source_code_location(this, self, file, lineNumber):
        _otf2.GlobalDefWriter_WriteSourceCodeLocation(this._handle, self, file, lineNumber)

    def write_calling_context(this, self, region, sourceCodeLocation, parent):
        _otf2.GlobalDefWriter_WriteCallingContext(this._handle, self, region, sourceCodeLocation, parent)

    def write_calling_context_property(this, callingContext, name, type, value):
        _otf2.GlobalDefWriter_WriteCallingContextProperty(this._handle, callingContext, name, type, value)

    def write_interrupt_generator(this, self, name, interruptGeneratorMode, base, exponent, period):
        _otf2.GlobalDefWriter_WriteInterruptGenerator(this._handle, self, name, interruptGeneratorMode, base, exponent, period)

    def write_io_file_property(this, ioFile, name, type, value):
        _otf2.GlobalDefWriter_WriteIoFileProperty(this._handle, ioFile, name, type, value)

    def write_io_regular_file(this, self, name, scope):
        _otf2.GlobalDefWriter_WriteIoRegularFile(this._handle, self, name, scope)

    def write_io_directory(this, self, name, scope):
        _otf2.GlobalDefWriter_WriteIoDirectory(this._handle, self, name, scope)

    def write_io_handle(this, self, name, file, ioParadigm, ioHandleFlags, comm, parent):
        _otf2.GlobalDefWriter_WriteIoHandle(this._handle, self, name, file, ioParadigm, ioHandleFlags, comm, parent)

    def write_io_pre_created_handle_state(this, ioHandle, mode, statusFlags):
        _otf2.GlobalDefWriter_WriteIoPreCreatedHandleState(this._handle, ioHandle, mode, statusFlags)

    def write_callpath_parameter(this, callpath, parameter, type, value):
        _otf2.GlobalDefWriter_WriteCallpathParameter(this._handle, callpath, parameter, type, value)

    def write_inter_comm(this, self, name, groupA, groupB, commonCommunicator, flags):
        _otf2.GlobalDefWriter_WriteInterComm(this._handle, self, name, groupA, groupB, commonCommunicator, flags)



# Generated by OTF2 Template Engine

import _otf2.GlobalDefWriter
from .definitions import *


class DefinitionWriter(object):
    """
        This class wraps the OTF2_GlobalDefWriter.

        Used internally only.
    """
    def __init__(self, archive):
        assert archive
        self._handle = _otf2.Archive_GetGlobalDefWriter(archive._handle)
        self._archive = archive

    def __enter__(self):
        assert self._handle
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        """Destructor.  Calls close()."""
        # The try/except block is in case this is called at program
        # exit time, when it's possible that globals have already been
        # deleted, and then the close() call might fail.  Since
        # there's nothing we can do about such failures and they annoy
        # the end users, we suppress the traceback.
        try:
            self.close()
        except:
            pass

    def close(self):
        if self._handle:
            try:
                _otf2.Archive_CloseGlobalDefWriter(self._archive._handle, self._handle)
            finally:
                self._archive = None
                self._handle = None

    @property
    def handle(self):
        """
            Gives access to the underlaying OTF2 global definition writer
        """
        if self._handle is None:
            raise TraceWriterError(
                "Trying to access handle, but GlobalDefWriter is already closed.")
        return self._handle

    @property
    def number_of_definitions(self):
        return _otf2.GlobalDefWriter.GetNumberOfDefinitions(self._handle)

    @property
    def number_of_locations(self):
        return _otf2.GlobalDefWriter.GetNumberOfLocations(self._handle)

    def write(self, definition, *args):
        {
@otf2 for def in defs|global_defs|exclude(ClockProperties):
            @@def.name@@: self.write_@@def.lower@@,
@otf2 endfor
        }[type(definition)](*args)
@otf2 for def in defs|global_defs:
@otf2  if def == IoParadigm:

    def write_io_paradigm(this, self, identification, name, ioParadigmClass, ioParadigmFlags, properties, types, values):
        _otf2.GlobalDefWriter_WriteIoParadigm(this._handle, self, identification, name,
                                              ioParadigmClass, ioParadigmFlags, properties, types, values)
@otf2  elif def == Group:

    def write_group(this, self, name, groupType, paradigm, groupFlags, members):
        _otf2.GlobalDefWriter_WriteGroup(this._handle, self, name, groupType, paradigm, groupFlags, members)
@otf2  elif def == MetricClass:

    def write_metric_class(this, self, metricMembers, metricOccurrence, recorderKind):
        _otf2.GlobalDefWriter_WriteMetricClass(this._handle, self, metricMembers, metricOccurrence, recorderKind)
@otf2  elif def == CartTopology:

    def write_cart_topology(this, self, name, communicator, cartDimensions):
        _otf2.GlobalDefWriter_WriteCartTopology(this._handle, self, name, communicator, cartDimensions)
@otf2  elif def == CartCoordinate:

    def write_cart_coordinate(this, cartTopology, rank, coordinates):
        _otf2.GlobalDefWriter_WriteCartCoordinate(this._handle, cartTopology, rank, coordinates)
@otf2  else:

    def write_@@def.lower@@(this@@def.py_funcargs()@@):
        _otf2.GlobalDefWriter_Write@@def.name@@(this._handle@@def.py_funcargs()@@)
@otf2  endif
@otf2 endfor



import ctypes

from .Config import conf, str_decode, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import TimeStamp, CallbackCode
from .Marker import MarkerRef, MarkerScope, MarkerSeverity
from .Callbacks import callback_wrapper


class MarkerReaderCallbacks(ctypes.Structure):
    pass

def MarkerReaderCallbacks_New():
    c_New = conf.lib.OTF2_MarkerReaderCallbacks_New
    c_New.argtypes = []
    c_New.restype = ctypes.POINTER(MarkerReaderCallbacks)
    return c_New()

def MarkerReaderCallbacks_Delete(markerReaderCallbacks):
    c_Delete = conf.lib.OTF2_MarkerReaderCallbacks_Delete
    c_Delete.argtype = [ ctypes.POINTER(MarkerReaderCallbacks) ]
    c_Delete.restype = None
    return c_Delete(markerReaderCallbacks)

def MarkerReaderCallbacks_Clear(markerReaderCallbacks):
    c_Clear = conf.lib.OTF2_MarkerReaderCallbacks_Clear
    c_Clear.argtypes = [ ctypes.POINTER(MarkerReaderCallbacks) ]
    c_Clear.restype = None
    return c_Clear(markerReaderCallbacks)

def _callback_wrapper(type, func, convert_args=None):
    def wrapper(userData, *args):
        if userData:
            py_userData = ctypes.cast(userData, ctypes.py_object).value
        else:
            py_userData = None
        try:
            if convert_args is not None:
                args = convert_args(*args)
            ret = func(py_userData, *args)
            if ret is None:
                ret = CALLBACK_SUCCESS
        except:
            sys.stderr.write("An unhandled python exception has occurred in an "
                             "OTF2_MarkerReaderCallback:\n")
            sys.stderr.write(traceback.format_exc())
            ret = CALLBACK_ERROR
        return ret.value
    return callback_wrapper(func, wrapper, type)

_MarkerReaderCallback_FP_Unknown = ctypes.CFUNCTYPE(CallbackCode, ctypes.c_void_p)

_MarkerReaderCallback_FP_DefMarker = ctypes.CFUNCTYPE(CallbackCode, ctypes.c_void_p, MarkerRef, ctypes.c_char_p, ctypes.c_char_p, MarkerSeverity)

_MarkerReaderCallback_FP_Marker = ctypes.CFUNCTYPE(CallbackCode, ctypes.c_void_p, TimeStamp, TimeStamp, MarkerRef, MarkerScope, ctypes.c_uint64, ctypes.c_char_p)

def MarkerReaderCallbacks_SetUnknownCallback(markerReaderCallbacks, unknownCallback):
    c_SetUnknownCallback = conf.lib.OTF2_MarkerReaderCallbacks_SetUnknownCallback
    c_SetUnknownCallback.argtypes = [ ctypes.POINTER(MarkerReaderCallbacks), _MarkerReaderCallback_FP_Unknown ]
    c_SetUnknownCallback.restype = ErrorCode
    c_SetUnknownCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_MarkerReaderCallback_FP_Unknown, unknownCallback)
    c_SetUnknownCallback(markerReaderCallbacks, wrapped_callback)

def MarkerReaderCallbacks_SetDefMarkerCallback(markerReaderCallbacks, defMarkerCallback):
    c_SetDefMarkerCallback = conf.lib.OTF2_MarkerReaderCallbacks_SetDefMarkerCallback
    c_SetDefMarkerCallback.argtypes = [ ctypes.POINTER(MarkerReaderCallbacks), _MarkerReaderCallback_FP_DefMarker ]
    c_SetDefMarkerCallback.restype = ErrorCode
    c_SetDefMarkerCallback.errcheck = HandleErrorCode
    def convert_string(ref, group, category, severity):
        return ref, str_decode(group), str_decode(category), severity
    wrapped_callback = _callback_wrapper(_MarkerReaderCallback_FP_DefMarker, defMarkerCallback, convert_string)
    c_SetDefMarkerCallback(markerReaderCallbacks, wrapped_callback)

def MarkerReaderCallbacks_SetMarkerCallback(markerReaderCallbacks, markerCallback):
    c_SetMarkerCallback = conf.lib.OTF2_MarkerReaderCallbacks_SetMarkerCallback
    c_SetMarkerCallback.argtypes = [ ctypes.POINTER(MarkerReaderCallbacks), _MarkerReaderCallback_FP_Marker ]
    c_SetMarkerCallback.restype = ErrorCode
    c_SetMarkerCallback.errcheck = HandleErrorCode
    def convert_string(ref, scope, scope_ref, text):
        return ref, scope, scope_ref, str_decode(text)
    wrapped_callback = _callback_wrapper(_MarkerReaderCallback_FP_Marker, markerCallback, convert_string)
    c_SetMarkerCallback(markerReaderCallbacks, wrapped_callback)

__all__ = [
    'MarkerReaderCallbacks',
    'MarkerReaderCallbacks_New',
    'MarkerReaderCallbacks_Delete',
    'MarkerReaderCallbacks_Clear',
    'MarkerReaderCallbacks_SetUnknownCallback',
    'MarkerReaderCallbacks_SetDefMarkerCallback',
    'MarkerReaderCallbacks_SetMarkerCallback'
]



import ctypes
import importlib.machinery
import os.path
from os import environ

from collections import namedtuple

Conf = namedtuple('Conf', 'lib')


def is_sphinx_run():
    try:
        return environ["OTF2_PYTHON_SPHINX_GENERATOR_RUN"] == "yes"
    except:
        return False


def lib_path():
    libname = '_otf2'
    for extension in importlib.machinery.EXTENSION_SUFFIXES:
        full_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), libname + extension)
        if os.path.isfile(full_path):
            return full_path
    raise FileNotFoundError("Could not find otf2 library.")


def lib():
    if is_sphinx_run():
        return None
    libhandle = ctypes.cdll.LoadLibrary(lib_path())
    return Conf(libhandle)


conf = lib()

# Performance note:
# str.encode('utf-8'), seems to be slightly faster then Codec.encode()[0]
# str.encode() using the default argument 'utf-8' may be slightly faster
# but is more confusing for encoding, there might be optimization potential
# by avoiding the intermediate byte object and see if a sub-class of
# c_char_p could be a bytes-like object itself
encoding = 'utf-8'

class StrParam(object):
    """
    argtype for ctypes to be used for string pointers
    Python3 variant, encodes the unicode to binary
    """
    @classmethod
    def from_param(cls, value):
        if value is None:
            return None
        assert isinstance(value, str)
        return value.encode(encoding)

def str_errcheck_nullable(result, func, arguments):
    """
    ctypes errcheck method for strings.
    NULL / None is ok.
    Use in combination with restype = ctypes.c_char_p
    Python3 variant, decodes the binary string to unicode
    """
    if result is None:
        return None
    assert isinstance(result, bytes)
    return result.decode(encoding)

def str_decode(otf2_str):
    """
    Decode a string from otf2 / ctypes.
    Asserts the input is a valid string or NULL.
    Python3 variant, decodes the binary string to unicode
    """
    if otf2_str is None:
        return None
    assert isinstance(otf2_str, bytes)
    return otf2_str.decode(encoding)


class EnumBase(object):
    def __eq__(self, other):
        if type(self) != type(other):
            raise TypeError("Trying to compare different enums.")
        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        try:
            return "{}.{}".format(type(self).__name__, type(self)._names[self.value])
        except AttributeError:
            return "{}({})".format(type(self).__name__, self.value)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((type(self), self.value))

    @classmethod
    def _construct(cls, value, name):
        assert type(name) is str
        assert name != ""
        if not hasattr(cls, "_names"):
            cls._names = dict()
        cls._names[value] = name
        obj = cls(value)
        setattr(cls, name, obj)
        return obj


def _same_type(a, b):
    assert type(a) == type(b)
    return type(a)


class EnumBitset(EnumBase):
    def __or__(self, other):
        return _same_type(self, other)(self.value | other.value)

    def __and__(self, other):
        return _same_type(self, other)(self.value & other.value)

    def __xor__(self, other):
        return _same_type(self, other)(self.value ^ other.value)

    def __invert__(self):
        return type(self)(~self.value)


def errcheck_null_to_none(result, func, arguments):
    if result:
        return result
    return None


def free(value):
    c_free = conf.lib.otf2_python_free
    c_free.argtypes = [ ctypes.c_void_p ]
    c_free.restype = None
    c_free(value)
